import abc
import uuid
from dataclasses import dataclass
from typing import List

import numpy as np
from scipy.spatial.transform import Rotation

from .. import config
from .._util import rotation_matrix_from_vectors
from ..config.types import ndarray
from ..exceptions import EmptySelectionError
from ..morphologies import MorphologySet, RotationSet
from ..profiling import node_meter
from ..storage._files import NrrdDependencyNode
from ..topology.partition import Partition
from .indicator import PlacementIndications


@dataclass
class DistributionContext:
    indicator: PlacementIndications
    partitions: List[Partition]


@config.dynamic(attr_name="strategy", required=True)
class Distributor(abc.ABC):
    def __init_subclass__(cls, **kwargs):
        super(cls, cls).__init_subclass__(**kwargs)
        # Decorate subclasses to measure performance
        node_meter("distribute")(cls)

    @abc.abstractmethod
    def distribute(self, positions, context):
        """
        Is called to distribute cell properties.

        :param partitions: The partitions the cells were placed in.
        :type context: Additional context information such as the placement indications
          and partitions.
        :type context: ~bsb.placement.distributor.DistributionContext
        :returns: An array with the property data
        :rtype: numpy.ndarray
        """
        pass


@config.dynamic(
    attr_name="strategy", required=False, default="random", auto_classmap=True
)
class MorphologyDistributor(Distributor):
    may_be_empty = config.attr(type=bool, default=False)

    @abc.abstractmethod
    def distribute(self, positions, morphologies, context):
        """
        Is called to distribute cell morphologies and optionally rotations.

        :param positions: Placed positions under consideration
        :type positions: numpy.ndarray
        :param morphologies: The template morphology loaders. You can decide to use them
          and/or generate new ones in the MorphologySet that you produce. If you produce
          any new morphologies, don't forget to encapsulate them in a
          :class:`~bsb.storage.interfaces.StoredMorphology` loader, or better yet, use
          the :class:`~bsb.placement.distributor.MorphologyGenerator`.
        :param context: The placement indicator and partitions.
        :type context: ~bsb.placement.distributor.DistributionContext
        :returns: A MorphologySet with assigned morphologies, and optionally a RotationSet
        :rtype: Union[~bsb.morphologies.MorphologySet, Tuple[
          ~bsb.morphologies.MorphologySet, ~bsb.morphologies.RotationSet]]
        """
        pass


@config.node
class RandomMorphologies(MorphologyDistributor, classmap_entry="random"):
    """
    Distributes selected morphologies randomly without rotating them.

    .. code-block:: json

      { "placement": { "place_XY": {
        "distribute": {
            "morphologies": {"strategy": "random"}
        }
      }}}
    """

    may_be_empty = config.provide(False)

    def distribute(self, positions, morphologies, context):
        """
        Uses the morphology selection indicators to select morphologies and
        returns a MorphologySet of randomly assigned morphologies
        """
        return np.random.default_rng().integers(len(morphologies), size=len(positions))


@config.node
class RoundRobinMorphologies(MorphologyDistributor, classmap_entry="roundrobin"):
    """
    Distributes selected morphologies round robin, values are looped and assigned one by
    one in order.

    .. code-block:: json

      { "placement": { "place_XY": {
        "distribute": {
            "morphologies": {"strategy": "roundrobin"}
        }
      }}}
    """

    may_be_empty = config.provide(False)

    def distribute(self, positions, morphologies, context):
        ll = len(morphologies)
        lp = len(positions)
        return np.tile(np.arange(ll), lp // ll + 1)[:lp]


@config.node
class MorphologyGenerator(MorphologyDistributor, classmap_entry=None):
    """
    Special case of the morphology distributor that provides extra convenience when
    generating new morphologies.
    """

    may_be_empty = config.attr(type=bool, default=True)

    def __init_subclass__(cls, **kwargs):
        super(cls, cls).__init_subclass__(**kwargs)
        # Decorate subclasses to measure performance
        node_meter("generate")(cls)

    def distribute(self, positions, morphologies, context):
        pass

    @abc.abstractmethod
    def generate(self, positions, morphologies, context):
        pass


@config.dynamic(attr_name="strategy", required=False, default="none", auto_classmap=True)
class RotationDistributor(Distributor):
    """
    Rotates everything by nothing!
    """

    @abc.abstractmethod
    def distribute(self, positions, context):
        pass


@config.node
class ExplicitNoRotations(RotationDistributor, classmap_entry="explicitly_none"):
    def distribute(self, positions, context):
        return np.zeros((len(positions), 3))


# Sentinel mixin class to tag RotationDistributor as being overridable by morphology
# distributor.
class Implicit:
    pass


@config.node
class ImplicitNoRotations(ExplicitNoRotations, Implicit, classmap_entry="none"):
    pass


@config.node
class RandomRotations(RotationDistributor, classmap_entry="random"):
    def distribute(self, positions, context):
        return np.random.rand(len(positions), 3) * 360


@config.node
class VolumetricRotations(RotationDistributor, classmap_entry="orientation_field"):
    orientation_path = config.attr(required=True, type=NrrdDependencyNode)
    """Path to the nrrd file containing the volumetric orientation field. It provides a rotation 
    for each voxel considered. Its shape should be (3, L, W, D) where L, W and D are the sizes of 
    the field."""
    orientation_resolution = config.attr(required=False, default=25.0, type=float)
    """Voxel size resolution of the orientation field.
    """
    default_vector = config.attr(
        required=False,
        default=lambda: np.array([0.0, -1.0, 0.0]),
        call_default=True,
        type=ndarray(),
    )
    """Default orientation vector of each position.
    """
    space_origin = config.attr(
        required=False,
        default=lambda: np.array([0.0, 0.0, 0.0]),
        call_default=True,
        type=ndarray(),
    )
    """Origin point for the orientation field.
    """

    def distribute(self, positions, context):
        """
        Rotates according to a volumetric orientation field of specific resolution.
        For each position, find the equivalent voxel in the volumetric orientation field and apply
        the rotation from the default_vector to the corresponding orientation vector.
        Positions outside the orientation field will not be rotated.

        :param positions: Placed positions under consideration. Its shape is (N, 3) where N is the
            number of positions.
        :param context: The placement indicator and partitions.
        :type context: ~bsb.placement.distributor.DistributionContext
        :returns: A RotationSet object containing the 3D Euler angles in degrees for the rotation
            of each position.
        :rtype: RotationSet
        """

        orientation_field = self.orientation_path.load_object()
        voxel_pos = np.asarray(
            np.floor((positions - self.space_origin) / self.orientation_resolution),
            dtype=int,
        )

        # filter for positions inside the orientation field.
        filter_inside = (
            np.all(voxel_pos >= 0, axis=1)
            * (voxel_pos[:, 0] < orientation_field.shape[1])
            * (voxel_pos[:, 1] < orientation_field.shape[2])
            * (voxel_pos[:, 2] < orientation_field.shape[3])
        )

        # By default, positions outside the field should not rotate.
        # So their target orientation vector will be set to the default_vector,
        # from which the rotation is processed.
        orientations = np.full((positions.shape[0], 3), self.default_vector, dtype=float)
        # Expected orientation_field shape is (3, L, W, D) where L, W and D are the sizes
        # of the field. Here we want to filter on the space dimensions, so we move the axes.
        if filter_inside.any():
            orientations[filter_inside] = np.moveaxis(orientation_field, 0, -1)[
                voxel_pos[filter_inside, 0],
                voxel_pos[filter_inside, 1],
                voxel_pos[filter_inside, 2],
            ]
            orientations[
                np.isnan(orientations).any(axis=1) + ~orientations.any(axis=1)
            ] = self.default_vector

        return RotationSet(
            Rotation.from_matrix(
                rotation_matrix_from_vectors(self.default_vector, v)
            ).as_euler("xyz", degrees=True)
            for v in orientations
        )


@config.node
class DistributorsNode:
    morphologies: MorphologyDistributor = config.attr(
        type=MorphologyDistributor, default=dict, call_default=True
    )
    rotations: RotationDistributor = config.attr(
        type=RotationDistributor, default=dict, call_default=True
    )
    properties: dict[Distributor] = config.catch_all(type=Distributor)

    def __call__(self, key, partitions, indicator, positions, loaders=None):
        context = DistributionContext(indicator, partitions)
        if key == "morphologies":
            distributor = getattr(self, key)
            if hasattr(distributor, "generate"):
                distribute = distributor.generate
            else:
                distribute = distributor.distribute
            values = distribute(positions, loaders, context)
            if isinstance(values, tuple):
                # Check for accidental tuple return values. If you return a 2 sized
                # tuple you're still fucked, but the Gods must truly hate you.
                try:
                    morphologies, rotations = values
                except TypeError:
                    raise ValueError(
                        "Morphology distributors may only return tuples when they are"
                        + " to be unpacked as (morphologies, rotations)"
                    ) from None
            else:
                values = (values, None)
            return values
        elif key == "rotations":
            distribute = getattr(self, key).distribute
        else:
            distribute = self.properties[key].distribute
        return distribute(positions, context)

    def _curry(self, partitions, indicator, positions, loaders=None):
        def curried(key):
            return self(key, partitions, indicator, positions, loaders)

        return curried

    def _specials(self, partitions, indicator, positions):
        sel = indicator.assert_indication("morphologies")
        loaders = self.scaffold.storage.morphologies.select(*sel)
        if not loaders and not self.morphologies.may_be_empty:
            raise EmptySelectionError(
                f"Given {len(sel)} selectors: did not find any suitable morphologies", sel
            )
        distr = self._curry(partitions, indicator, positions, loaders)
        morphologies, rotations = distr("morphologies")
        if morphologies is not None and (
            rotations is None or not isinstance(self.rotations, Implicit)
        ):
            # If a RotationDistributor is not explicitly marked as `Implicit`, it
            # overrides the MorphologyDistributor's implicit rotations.
            rotations = distr("rotations")
        if hasattr(self.morphologies, "generate"):
            prefix = self._config_parent.name
            generated = {}
            indices = []
            # Get all the unique morphology objects from the return value and map to them
            for m in morphologies:
                idx = generated.setdefault(m, len(generated))
                indices.append(idx)
            mr = self.scaffold.morphologies
            uid = uuid.uuid4()
            loaders = []
            all_meta = {}
            # Save morphologies one by one
            for gen_morpho, i in generated.items():
                name = f"{prefix}-{uid}-{i}"
                # `update_metadata=False` so morphology is incompletely stored.
                saved = mr.save(name, gen_morpho, update_meta=False)
                all_meta[name] = saved.get_meta()
                loaders.append(saved)
            # Finish morphologies by bulk saving their collective metadata
            mr.update_all_meta(all_meta)
            morphologies = MorphologySet(loaders, indices)
        if not isinstance(morphologies, MorphologySet) and morphologies is not None:
            morphologies = MorphologySet(loaders, morphologies)

        return morphologies, rotations

    def _has_mdistr(self):
        # This function checks if this distributor node has specified a morpho distributor
        return self.__class__.morphologies.is_dirty(self)

    def _has_rdistr(self):
        # This function checks if this distributor node has specified a rotation distributor
        return self.__class__.rotations.is_dirty(self)


__all__ = [
    "DistributionContext",
    "Distributor",
    "DistributorsNode",
    "ExplicitNoRotations",
    "Implicit",
    "ImplicitNoRotations",
    "MorphologyDistributor",
    "MorphologyGenerator",
    "RandomMorphologies",
    "RandomRotations",
    "RotationDistributor",
    "RoundRobinMorphologies",
    "VolumetricRotations",
]



import abc
import concurrent
import re
import tempfile
import typing
import urllib
import warnings
from concurrent.futures import ThreadPoolExecutor

import requests

from .. import config
from ..config import types
from ..config._attrs import cfglist
from ..exceptions import MissingMorphologyError, SelectorError
from ..services import MPI
from . import Morphology

if typing.TYPE_CHECKING:
    from ..core import Scaffold


@config.dynamic(
    attr_name="select",
    auto_classmap=True,
    required=False,
    default="by_name",
)
class MorphologySelector(abc.ABC):
    scaffold: "Scaffold"

    @abc.abstractmethod
    def validate(self, all_morphos):
        pass

    @abc.abstractmethod
    def pick(self, morphology):
        pass


@config.node
class NameSelector(MorphologySelector, classmap_entry="by_name"):
    names: cfglist[str] = config.list(type=str, required=types.shortform())

    def __init__(self, name=None, /, **kwargs):
        if name is not None:
            self.names = [name]

    def __inv__(self):
        if self._config_pos_init:
            return self.names[0]
        return self.__tree__()

    def _cache_patterns(self):
        self._pnames = {n: n.replace("*", r".*").replace("|", "\\|") for n in self.names}
        self._patterns = {n: re.compile(f"^{pat}$") for n, pat in self._pnames.items()}
        self._empty = not self.names
        self._match = re.compile(f"^({'|'.join(self._pnames.values())})$")

    def validate(self, all_morphos):
        self._cache_patterns()
        repo_names = {m.get_meta()["name"] for m in all_morphos}
        missing = [
            n
            for n, pat in self._patterns.items()
            if not any(pat.match(rn) for rn in repo_names)
        ]
        if missing:
            err = "Morphology repository misses the following morphologies"
            if self._config_parent is not None:
                node = self._config_parent._config_parent
                err += f" required by {node.get_node_name()}"
            err += f": {', '.join(missing)}"
            raise MissingMorphologyError(err)

    def pick(self, morphology):
        self._cache_patterns()
        return (
            not self._empty
            and self._match.match(morphology.get_meta()["name"]) is not None
        )


@config.node
class NeuroMorphoSelector(NameSelector, classmap_entry="from_neuromorpho"):
    _url = "https://neuromorpho.org/"
    _meta = "api/neuron/select?q=neuron_name:"
    _files = "dableFiles/"

    def __boot__(self):
        if self.scaffold.is_main_process():
            try:
                morphos = self._scrape_nm(self.names)
            except:
                MPI.barrier()
                raise
            for name, morpho in morphos.items():
                self.scaffold.morphologies.save(name, morpho, overwrite=True)
        MPI.barrier()

    @classmethod
    def _swc_url(cls, archive, name):
        return f"{cls._url}{cls._files}{urllib.parse.quote(archive.lower())}/CNG%20version/{name}.CNG.swc"

    @classmethod
    def _scrape_nm(cls, names):
        # Weak DH key on neuromorpho.org
        # https://stackoverflow.com/questions/38015537/python-requests-exceptions-sslerror-dh-key-too-small
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ":HIGH:!DH:!aNULL"
        try:
            requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += (
                ":HIGH:!DH:!aNULL"
            )
        except AttributeError:
            # no pyopenssl support used / needed / available
            pass
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with ThreadPoolExecutor() as executor:
                # Certificate issues with neuromorpho --> verify=False
                res = requests.get(cls._url + cls._meta + ",".join(names), verify=False)
                if res.status_code == 404:
                    raise SelectorError(f"'{names[0]}' is not a valid NeuroMorpho name.")
                elif res.status_code != 200:
                    raise SelectorError("NeuroMorpho API error: " + res.message)
                metas = {n: None for n in names}
                for meta in res.json()["_embedded"]["neuronResources"]:
                    del meta["_links"]
                    metas[meta["neuron_name"]] = meta
                missing = [name for name, meta in metas.items() if meta is None]
                if missing:
                    raise SelectorError(
                        ", ".join(f"'{n}'" for n in missing)
                        + " are not valid NeuroMorpho names."
                    )
                swc_urls = {n: cls._swc_url(metas[n]["archive"], n) for n in names}
                req = lambda n: requests.get(swc_urls[n], verify=False)
                sub = lambda n: (executor.submit(req, n), n)
                futures = dict(map(sub, names))
                morphos = {n: None for n in names}
                with tempfile.TemporaryDirectory() as tempdir:
                    for future in concurrent.futures.as_completed(futures.keys()):
                        name = futures[future]
                        path = tempdir + f"/{name}.swc"
                        with open(path, "w") as f:
                            f.write(future.result().text)
                        morphos[name] = Morphology.from_swc(path, meta=metas[name])
                missing = [name for name, m in morphos.items() if m is None]
                if missing:  # pragma: nocover
                    raise SelectorError(
                        "Downloading NeuroMorpho failed for "
                        + ", ".join(f"'{n}'" for n in missing)
                        + "."
                    )
                return morphos


__all__ = ["MorphologySelector", "NameSelector", "NeuroMorphoSelector"]



import abc
import functools

from ..exceptions import PluginError


class ConfigurationParser(abc.ABC):
    @abc.abstractmethod
    def parse(self, content, path=None):
        pass

    @abc.abstractmethod
    def generate(self, tree, pretty=False):
        pass


@functools.cache
def get_configuration_parser_classes():
    from ..plugins import discover

    return discover("config.parsers")


def get_configuration_parser(parser, **kwargs):
    """
    Create an instance of a configuration parser that can parse configuration
    strings into configuration trees, or serialize trees into strings.

    Configuration trees can be cast into Configuration objects.
    """
    parsers = get_configuration_parser_classes()
    if parser not in parsers:
        raise PluginError(f"Configuration parser '{parser}' not found")
    return parsers[parser](**kwargs)


__all__ = [
    "ConfigurationParser",
    "get_configuration_parser",
    "get_configuration_parser_classes",
]



import functools
import itertools

import numpy as np

from .exceptions import EmptyVoxelSetError
from .trees import BoxTree


class VoxelData(np.ndarray):
    """
    Chunk identifier, consisting of chunk coordinates and size.
    """

    def __new__(cls, data, keys=None):
        if data.ndim < 2:
            return super().__new__(np.ndarray, data.shape, dtype=object)
        obj = super().__new__(cls, data.shape, dtype=object)
        obj[:] = data
        if keys is not None:
            keys = [str(k) for k in keys]
            if len(set(keys)) != len(keys):
                raise ValueError("Data keys must be unique")
            if len(keys) != data.shape[1]:
                raise ValueError("Amount of data keys must match amount of data columns")
            obj._keys = keys
        else:
            obj._keys = []
        return obj

    def __getitem__(self, index):
        index, keys = self._rewrite_index(index)
        vd = super().__getitem__(index)
        if isinstance(vd, VoxelData):
            if len(keys) > 0 and len(vd) != vd.size / len(keys):
                vd = vd.reshape(-1, len(keys))
            vd._keys = keys
        return vd

    def __array_finalize__(self, obj):
        if obj is not None:
            self._keys = []

    @property
    def keys(self):
        """
        Returns the keys, or column labels, associated to each data column.
        """
        return self._keys.copy()

    def copy(self):
        """
        Return a new copy of the voxel data
        """
        new = super().copy()
        new._keys = self._keys.copy()
        return new

    def _split_index(self, index):
        try:
            if isinstance(index, tuple):
                cols = [self._keys.index(idx) for idx in index if isinstance(idx, str)]
                keys = [self._keys[c] for c in cols]
                index = tuple(idx for idx in index if not isinstance(idx, str))
            elif isinstance(index, str):
                cols = [self._keys.index(index)]
                keys = [self._keys[cols[0]]]
                index = (slice(None),)
            else:
                index = (index,)
                cols = None
                keys = getattr(self, "_keys", [])
        except ValueError as e:
            key = str(e).split("'")[1]
            raise IndexError(f"Voxel data key '{key}' does not exist.") from None
        return index, cols, keys

    def _rewrite_index(self, index):
        index, cols, keys = self._split_index(index)
        if cols:
            return (*index, cols), keys
        else:
            return index, keys


class VoxelSet:
    def __init__(self, voxels, size, data=None, data_keys=None, irregular=False):
        """
        Constructs a voxel set from the given voxel indices or coordinates.

        :param voxels: The spatial coordinates, or grid indices of each voxel. Nx3 matrix
        :type voxels: numpy.ndarray
        :param size: The global or individual size of the voxels. If it has 2 dimensions
          it needs to have the same length as `voxels`, and will be used as individual
          voxels.
        :type size: numpy.ndarray
        :param data:

        .. warning::

            If :class:`numpy.ndarray` are passed, they will not be copied in order to save
            memory and time. You may accidentally change a voxelset if you later change
            the same array.
        """
        voxels = np.array(voxels, copy=False)
        voxel_size = np.array(size, copy=False)
        if voxels.dtype.name == "object":
            raise ValueError("Couldn't convert given `voxels` to a voxel matrix")
        if voxels.ndim != 2:
            if not len(voxels):
                # Try some massaging in case of empty arrays
                voxels = voxels.reshape(-1, 3)
                if not len(voxel_size):
                    voxel_size = voxel_size.reshape(-1, 3)
            else:
                raise ValueError("`voxels` needs to be convertable to a 2D matrix")
        if voxels.ndim == 2 and voxels.shape[1] != 3:
            raise ValueError("`voxels` needs to have 3 columns, 1 for each spatial dim.")
        if not _is_broadcastable(voxels.shape, voxel_size.shape):
            raise ValueError(
                f"Shape {voxel_size.shape} of `size` is"
                + f" invalid for voxel shape {voxels.shape}"
            )
        if data is not None:
            if isinstance(data, VoxelData):
                if data_keys is None:
                    self._data = data
                else:
                    self._data = VoxelData(data, keys=data_keys)
            else:
                data = np.array(data, copy=False)
                if data.ndim < 2:
                    cols = len(data_keys) if data_keys else 1
                    data = data.reshape(-1, cols)
                self._data = VoxelData(data, keys=data_keys)
            if len(self._data) != len(voxels):
                raise ValueError("`voxels` and `data` length unequal.")
        else:
            self._data = None

        if not len(voxel_size.shape):
            self._cubic = True
        if voxel_size.ndim > 1:
            if voxel_size.size != voxels.size:
                raise ValueError("`voxels` and `size` length unequal.")
            # Voxels given in spatial coords with individual size
            self._sizes = voxel_size
            self._coords = voxels
            self._regular = False
        elif irregular:
            # Voxels given in spatial coords but of equal size
            self._size = voxel_size
            self._coords = voxels
            self._regular = False
        else:
            # Voxels given in index coords
            self._size = voxel_size
            self._indices = np.array(voxels, copy=False, dtype=int)
            self._regular = True

    def __iter__(self):
        return iter(self.get_raw(copy=False))

    def __array__(self):
        return self.get_raw(copy=False)

    def __len__(self):
        return len(self.get_raw(copy=False))

    def __getitem__(self, index):
        if self.has_data:
            data = self._data[index]
            index, _, _ = self._data._split_index(index)
        else:
            data, _ = None, None
        if isinstance(index, tuple) and len(index) > 1:
            raise IndexError("Too many indices for VoxelSet, maximum 1.")
        voxels = self.get_raw(copy=False)[index]
        if self._single_size:
            voxel_size = self._size.copy()
        else:
            voxel_size = self._sizes[index]
        if voxels.ndim == 1:
            voxels = voxels.reshape(-1, 3)
        return VoxelSet(voxels, voxel_size, data)

    def __getattr__(self, key):
        if key in self._data._keys:
            return self.get_data(key)
        else:
            return super().__getattribute__(key)

    def __str__(self):
        cls = type(self)
        obj = f"<{cls.__module__}.{cls.__name__} object at {hex(id(self))}>"
        if self.is_empty:
            insert = "[EMPTY] "
        else:
            insert = f"with {len(self)} voxels from "
            insert += f"{tuple(self.bounds[0])} to {tuple(self.bounds[1])}, "
            if self.regular:
                insert += f"same size {self.size}, "
            else:
                insert += "individual sizes, "
            if self.has_data:
                if self._data.keys:
                    insert += f"with keyed data ({', '.join(self._data.keys)}) "
                else:
                    insert += f"with {self._data.shape[-1]} data columns "
            else:
                insert += "without data "
        return obj.replace("at 0x", insert + "at 0x")

    __repr__ = __str__

    def __bool__(self):
        return not self.is_empty

    @property
    def is_empty(self):
        """
        Whether the set contain any voxels

        :rtype: bool
        """
        return not len(self)

    @property
    def has_data(self):
        """
        Whether the set has any data associated to the voxels

        :rtype: bool
        """
        return self._data is not None

    @property
    def regular(self):
        """
        Whether the voxels are placed on a regular grid.

        :rtype: bool
        """
        return self._regular

    @property
    def of_equal_size(self):
        return self._single_size or len(np.unique(self._sizes, axis=0)) < 2

    @property
    def equilateral(self):
        """
        Whether all sides of all voxels have the same lengths.

        :rtype: bool
        """
        return np.unique(self.get_size(copy=False)).shape == (1,)

    @property
    def size(self):
        """
        The size of the voxels. When it is 0D or 1D it counts as the size for all voxels,
        if it is 2D it is 1 an individual size per voxel.

        :rtype: numpy.ndarray
        """
        return self.get_size()

    @property
    def volume(self):
        if self._single_size:
            voxel_volume = np.abs(np.prod(self.get_size(copy=False) * np.ones(3)))
            return voxel_volume * len(self)
        else:
            return np.sum(np.abs(np.prod(self.get_size_matrix(copy=False), axis=1)))

    @property
    def data(self):
        """
        The size of the voxels. When it is 0D or 1D it counts as the size for all voxels,
        if it is 2D it is 1 an individual size per voxel.

        :rtype: Union[numpy.ndarray, None]
        """
        return self.get_data()

    @property
    def data_keys(self):
        return self._data.keys

    @property
    def raw(self):
        return self.get_raw()

    @property
    @functools.cache
    def bounds(self):
        """
        The minimum and maximum coordinates of this set.

        :rtype: tuple[numpy.ndarray, numpy.ndarray]
        """
        if self.is_empty:
            raise EmptyVoxelSetError("Empty VoxelSet has no bounds.")
        boxes = self.as_boxes()
        dims = boxes.shape[1] // 2
        return (
            np.min(boxes[:, :dims], axis=0),
            np.max(boxes[:, dims:], axis=0),
        )

    @classmethod
    def empty(cls, size=None):
        return cls(np.empty((0, 3)), np.empty((0, 3)))

    @classmethod
    def one(cls, ldc, mdc, data=None):
        ldc = np.array(ldc, copy=False).reshape(-1)
        mdc = np.array(mdc, copy=False).reshape(-1)
        if ldc.shape != (3,) or mdc.shape != (3,):
            raise ValueError(
                "Arguments to `VoxelSet.one` should be shape (3,) minmax coords of voxel."
            )
        return cls(np.array([ldc]), np.array([mdc - ldc]), np.array([data]))

    @classmethod
    def concatenate(cls, *sets):
        # Short circuit "stupid" concat requests
        if not sets:
            return cls.empty()
        elif len(sets) == 1:
            return sets[0].copy()

        primer = None
        # Check which sets we are concatenating, maybe we can keep them in reduced data
        # forms. If they don't line up, we expand and concatenate the expanded forms.
        if any(
            # `primer` is assigned the first non-empty set, all sizes must match sizes can
            # still be 0D, 1D or 2D, but if they're allclose broadcasted it is fine! :)
            not np.allclose(s.get_size(copy=False), primer.get_size(copy=False))
            for s in sets
            if s and (primer := primer or s)
        ):
            sizes = primer.get_size()
            if len(sizes.shape) > 1:
                # We happened to pick a VoxelSet that has a size matrix of equal sizes,
                # so we take the opportunity to reduce it.
                sizes = sizes[0]
            if len(sizes.shape) > 0 and np.allclose(sizes, sizes[0]):
                # Voxelset is actually even cubic regular!
                sizes = sizes[0]
            if all(s.regular for s in sets):
                # Index coords with same sizes can simply be stacked
                voxels = np.concatenate([s.get_raw(copy=False) for s in sets])
                irregular = False
            else:
                voxels = np.concatenate([s.as_spatial_coords(copy=False) for s in sets])
                irregular = True
        else:
            # We can't keep a single size, so expand into a matrix where needed and concat
            sizes = np.concatenate([s.get_size_matrix(copy=False) for s in sets])
            voxels = np.concatenate([s.as_spatial_coords(copy=False) for s in sets])
            irregular = True

        if any(s.has_data for s in sets):
            fillers = [s.get_data(copy=False) for s in sets]
            # Find all keys among data to concatenate
            all_keys = set(itertools.chain(*(f.keys for f in fillers if f is not None)))
            # Create an index for each key
            keys = [*sorted(set(all_keys), key=str)]
            # Allocate enough columns for all keys, or a data array with more unlabelled
            # columns than that.
            md = max(len(keys), *(f.shape[1] for f in fillers if f is not None))
            if not keys:
                keys = None
            elif md > len(keys):
                # Find and pad `keys` with labels for the extra numerical columns.
                extra = md - len(keys)
                new_nums = (s for c in itertools.count() if (s := str(c)) not in keys)
                keys.extend(itertools.islice(new_nums, extra))
                keys.extend(range(len(keys), md))
                keys = sorted(keys)
            ln = [len(s) for s in sets]
            data = np.empty((sum(ln), md), dtype=object)
            ptr = 0
            for len_, fill in zip(ln, fillers):
                if fill is not None:
                    if not fill.keys:
                        cols = slice(None, fill.shape[1])
                    else:
                        cols = [keys.index(key) for key in fill.keys]
                    data[ptr : (ptr + len_), cols] = fill
                    ptr += len_
        else:
            data = None
            keys = None
        return VoxelSet(voxels, sizes, data=data, data_keys=keys, irregular=irregular)

    def copy(self):
        if self.is_empty:
            return VoxelSet.empty()
        else:
            return VoxelSet(
                self.raw,
                self.get_size(copy=True),
                self.get_data(copy=True) if self.has_data else None,
                irregular=not self.regular,
            )

    def get_raw(self, copy=True):
        coords = self._indices if self.regular else self._coords
        if copy:
            coords = coords.copy()
        return coords

    def get_data(self, index=None, /, copy=True):
        if self.has_data:
            if index is not None:
                return self._data[index]
            else:
                return self._data.copy()
        else:
            return None

    def get_size(self, copy=True):
        if self._single_size:
            return np.array(self._size, copy=copy)
        else:
            return np.array(self._sizes, copy=copy)

    def get_size_matrix(self, copy=True):
        if self._single_size:
            size = np.ones(3) * self._size
            sizes = np.tile(size, (len(self.get_raw(copy=False)), 1))
        else:
            sizes = self._sizes
            if copy:
                sizes = sizes.copy()
        return sizes

    def as_spatial_coords(self, copy=True):
        if self.regular:
            coords = self._to_spatial_coords()
        else:
            coords = self._coords
            if copy:
                coords = coords.copy()
        return coords

    def as_boxes(self, cache=False):
        if cache:
            return self._boxes_cache()
        else:
            return self._boxes()

    def as_boxtree(self, cache=False):
        if cache:
            return self._boxtree_cache()
        else:
            return self._boxtree()

    def snap_to_grid(self, voxel_size, unique=False):
        if self.regular:
            grid = self._indices // _squash_zero(voxel_size / _squash_zero(self._size))
        else:
            grid = self._coords // _squash_zero(voxel_size)
        data = self._data
        if unique:
            if self.has_data:
                grid, id = np.unique(grid, return_index=True, axis=0)
                data = data[id]
            else:
                grid = np.unique(grid, axis=0)
        return VoxelSet(grid, voxel_size, data)

    def resize(self, size):
        val = np.array(size, copy=False)
        if val.dtype.name == "object":
            raise ValueError("Size must be number type")
        if val.ndim > 1:
            if len(val) != len(self):
                raise ValueError("Individual voxel sizes must match amount of voxels.")
            if self.regular:
                self._coords = self.as_spatial_coords()
                del self._indices
                self._regular = False
        self._size = size

    def crop(self, ldc, mdc):
        coords = self.as_spatial_coords(copy=False)
        inside = np.all(np.logical_and(ldc <= coords, coords < mdc), axis=1)
        return self[inside]

    def crop_chunk(self, chunk):
        return self.crop(chunk.ldc, chunk.mdc)

    @classmethod
    def fill(cls, positions, voxel_size, unique=True):
        return cls(positions, 0, irregular=True).snap_to_grid(voxel_size, unique=unique)

    def coordinates_of(self, positions):
        if not self.regular:
            raise ValueError("Cannot find a unique voxel index in irregular VoxelSet.")
        return positions // self.get_size()

    def index_of(self, positions):
        coords = self.coordinates_of(positions)
        map_ = {tuple(vox_coord): i for i, vox_coord in enumerate(self)}
        return np.array([map_.get(tuple(coord), np.nan) for coord in coords])

    def inside(self, positions):
        mask = np.zeros(len(positions), dtype=bool)
        ldc, mdc = self._box_bounds()
        for voxel in zip(ldc, mdc):
            mask |= np.all((positions >= ldc) & (positions < mdc), axis=1)
        return mask

    def unique(self):
        raise NotImplementedError("and another one")

    @property
    def _single_size(self):
        # One size fits all
        return hasattr(self, "_size")

    def _to_spatial_coords(self):
        return self._indices * self._size

    @functools.cache
    def _boxtree_cache(self):
        return self._boxtree()

    def _boxtree(self):
        return BoxTree(self.as_boxes())

    @functools.cache
    def _boxes_cache(self):
        return self._boxes()

    def _box_bounds(self):
        base = self.as_spatial_coords(copy=False)
        sizes = self.get_size(copy=False)
        shifted = base + sizes
        lt0 = sizes < 0
        if np.any(lt0):
            mdc = np.where(lt0, base, shifted)
            ldc = np.where(lt0, shifted, base)
        else:
            ldc = base
            mdc = shifted
        return ldc, mdc

    def _boxes(self):
        return np.column_stack(self._box_bounds())

    @classmethod
    def from_morphology(cls, morphology, estimate_n, with_data=True):
        meta = morphology.meta
        if "mdc" in meta and "ldc" in meta:
            ldc, mdc = meta["ldc"], meta["mdc"]
        else:
            ldc, mdc = morphology.bounds
        # Find a good distribution of amount of voxels per side
        size = mdc - ldc
        per_side = _eq_sides(size, estimate_n)
        voxel_size = size / per_side
        _squash_temp = _squash_zero(voxel_size)
        branch_vcs = [b.points // _squash_temp for b in morphology.branches]
        if with_data:
            voxel_reduce = {}
            for branch, point_vcs in enumerate(branch_vcs):
                for point, point_vc in enumerate(point_vcs):
                    voxel_reduce.setdefault(tuple(point_vc), []).append((branch, point))
            voxels = np.array(tuple(voxel_reduce.keys()))
            # Transfer the voxel data into an object array
            voxel_data_data = tuple(voxel_reduce.values())
            # We need a bit of a workaround so that numpy doesn't make a regular from the
            # `voxel_data_data` list of lists, when it has a matrix shape.
            voxel_data = np.empty(len(voxel_data_data), dtype=object)
            for i in range(len(voxel_data_data)):
                voxel_data[i] = voxel_data_data[i]
            return cls(voxels, voxel_size, data=voxel_data)
        else:
            voxels = np.unique(np.concatenate(branch_vcs), axis=0)
            return cls(voxels, voxel_size)


def _eq_sides(sides, n):
    zeros = np.isclose(sides, 0)
    if all(zeros):
        # An empty or 1 point morphology should make only 1 (empty) voxel
        return np.array([1])
    elif any(zeros):
        # Remove any zeros, by fixing their dimensions to 1 (zero-width) partition
        solution = np.ones(len(sides))
        solution[~zeros] = _eq_sides(sides[~zeros], n)
        return solution
    elif len(sides) == 1:
        # Only 1 dimension, only 1 solution: all voxels along that dimension.
        return np.array([n])

    # Use the relative magnitudes of each side
    norm = sides / max(sides)
    # Find out how many divisions each side should to form a grid with `n` rhomboids.
    per_side = norm * (n / np.prod(norm)) ** (1 / len(sides))
    # Divisions should be integers, and minimum 1
    solution = np.maximum(np.floor(per_side), 1)
    order = np.argsort(sides)
    smallest = order[0]
    if len(sides) > 2:
        # Because of the integer rounding the product isn't necesarily optimal, so we keep
        # the safest (smallest) value, and solve the problem again in 1 less dimension.
        solved = solution[smallest]
        look_for = n / solved
        others = sides[order[1:]]
        solution[order[1:]] = _eq_sides(others, look_for)
    else:
        # In the final 2-dimensional case the remainder of the division is rounded off
        # to the nearest integer, giving the smallest error on the product and final
        # number of rhomboids in the grid.
        largest = order[1]
        solution[largest] = round(n / solution[smallest])
    return solution


# https://stackoverflow.com/a/24769712/1016004
def _is_broadcastable(shape1, shape2):
    for a, b in zip(shape1[::-1], shape2[::-1]):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True


def _squash_zero(arr):
    return np.where(np.isclose(arr, 0), np.finfo(float).max, arr)


__all__ = ["BoxTree", "VoxelData", "VoxelSet"]



"""
Contains builtin commands.
"""

import itertools
from uuid import uuid4

import errr

from ..._options import ConfigOption
from ...config import parse_configuration_file
from ...core import Scaffold, from_storage
from ...exceptions import NodeNotFoundError
from ...option import BsbOption
from ...storage import open_storage
from . import BaseCommand


class XScale(BsbOption, name="x", cli=("x",), env=("BSB_CONFIG_NETWORK_X",)):
    pass


class YScale(BsbOption, name="y", cli=("y",), env=("BSB_CONFIG_NETWORK_Y",)):
    pass


class ZScale(BsbOption, name="z", cli=("z",), env=("BSB_CONFIG_NETWORK_Z",)):
    pass


class Skip(BsbOption, name="skip", cli=("skip",), env=("BSB_SELECTION_SKIP",), list=True):
    pass


class Only(BsbOption, name="only", cli=("only",), env=("BSB_SELECTION_ONLY",), list=True):
    pass


class Append(
    BsbOption, name="append", cli=("append", "a"), env=("BSB_APPEND_MODE",), flag=True
):
    pass


class Redo(BsbOption, name="redo", cli=("redo", "r"), env=("BSB_REDO_MODE",), flag=True):
    pass


class Clear(
    BsbOption, name="clear", cli=("clear", "w"), env=("BSB_CLEAR_MODE",), flag=True
):
    pass


class Output(BsbOption, name="output", cli=("output", "o"), env=("BSB_OUTPUT_FILE",)):
    pass


class SkipPlacement(
    BsbOption,
    name="skip_placement",
    cli=("np", "skip-placement"),
    env=("BSB_SKIP_PLACEMENT",),
    flag=True,
):
    pass


class SkipAfterPlacement(
    BsbOption,
    name="skip_after_placement",
    cli=("nap", "skip-after-placement"),
    env=("BSB_SKIP_AFTER_PLACEMENT",),
    flag=True,
):
    pass


class SkipConnectivity(
    BsbOption,
    name="skip_connectivity",
    cli=("nc", "skip-connectivity"),
    env=("BSB_SKIP_CONNECTIVITY",),
    flag=True,
):
    pass


class SkipAfterConnectivity(
    BsbOption,
    name="skip_after_connectivity",
    cli=("nac", "skip-after-connectivity"),
    env=("BSB_SKIP_AFTER_CONNECTIVITY",),
    flag=True,
):
    pass


class IgnoreErrors(
    BsbOption,
    name="ignore_errors",
    cli=("ignore", "ignore-errors"),
    env=("BSB_IGNORE_ERRORS",),
    flag=True,
):
    pass


def _flatten_arr_args(arr):
    if arr is None:
        return arr
    else:
        return list(itertools.chain.from_iterable(a.split(",") for a in arr))


class MakeConfigCommand(BaseCommand, name="make-config"):
    def handler(self, context):
        from ...config import copy_template

        args = context.arguments
        copy_template(args.template, args.output, path=args.path or ())

    def get_options(self):
        return {}

    def add_parser_arguments(self, parser):
        parser.add_argument("template", nargs="?", default="skeleton.json")
        parser.add_argument("output", nargs="?", default="network_configuration.json")
        parser.add_argument(
            "--path",
            help="Additional paths to search for config templates",
            action="extend",
            nargs="+",
            default=False,
        )


class BsbCompile(BaseCommand, name="compile"):
    def handler(self, context):
        cfg = parse_configuration_file(context.config)
        network = Scaffold(cfg)
        network.resize(context.x, context.y, context.z)
        network.compile(
            skip_placement=context.skip_placement,
            skip_after_placement=context.skip_after_placement,
            skip_connectivity=context.skip_connectivity,
            skip_after_connectivity=context.skip_after_connectivity,
            only=_flatten_arr_args(context.only),
            skip=_flatten_arr_args(context.skip),
            clear=context.clear,
            force=context.force,
            append=context.append,
            redo=context.redo,
            fail_fast=not context.ignore_errors,
        )

    def get_options(self):
        return {
            "x": XScale(),
            "y": YScale(),
            "z": ZScale(),
            "skip": Skip(),
            "only": Only(),
            "config": ConfigOption(positional=True),
            "no_placement": SkipPlacement(),
            "no_after_placement": SkipAfterPlacement(),
            "no_connectivity": SkipConnectivity(),
            "no_after_connectivity": SkipAfterConnectivity(),
            "append": Append(),
            "redo": Redo(),
            "clear": Clear(),
            "output": Output(),
            "ignore_errors": IgnoreErrors(),
        }

    def add_parser_arguments(self, parser):
        pass


class BsbReconfigure(BaseCommand, name="reconfigure"):
    def handler(self, context):
        cfg = parse_configuration_file(context.config)
        # Bootstrap the scaffold and clear the storage if not in append mode
        storage = open_storage(context.arguments.network)
        storage.store_active_config(cfg)

    def get_options(self):
        return {
            "config": ConfigOption(positional=True),
        }

    def add_parser_arguments(self, parser):
        parser.add_argument("network")


class BsbSimulate(BaseCommand, name="simulate"):
    def handler(self, context):
        network = from_storage(context.arguments.network)
        config_option = context.options["config"]
        sim_name = context.arguments.simulation
        extra_simulations = {}
        if config_option.is_set("cli"):
            extra_simulations = parse_configuration_file(context.config).simulations
            for name, sim in extra_simulations.items():
                if name not in network.simulations and name == sim_name:
                    network.simulations[sim_name] = sim
        try:
            result = network.run_simulation(sim_name)
        except NodeNotFoundError as e:
            append = ", " if len(network.simulations) else ""
            append += ", ".join(f"'{name}'" for name in extra_simulations.keys())
            errr.wrap(type(e), e, append=append)
        else:
            result.write(getattr(context.arguments, "output", f"{uuid4()}.nio"), "ow")

    def get_options(self):
        return {
            "skip": Skip(),
            "only": Only(),
        }

    def add_parser_arguments(self, parser):
        parser.add_argument("network")
        parser.add_argument("simulation")
        parser.add_argument("-o", "--output")


class CacheCommand(BaseCommand, name="cache"):  # pragma: nocover
    def handler(self, context):
        import shutil
        from datetime import datetime

        from ...storage._util import _cache_path

        if context.clear:
            shutil.rmtree(_cache_path)
            _cache_path.mkdir(parents=True, exist_ok=True)
            print("Cache cleared")
        else:
            _cache_path.mkdir(parents=True, exist_ok=True)
            files = [*_cache_path.iterdir()]
            maxlen = 5
            try:
                maxlen = max(maxlen, max(len(file.name) for file in files))
            except ValueError:
                print("Cache is empty")
            else:
                print(f"{'Files'.ljust(maxlen, ' ')}    Cached at\t\t\t    Size")
                total_mb = 0
                for f in files:
                    name = f.name.ljust(maxlen, " ")
                    stat = f.stat()
                    stamp = datetime.fromtimestamp(stat.st_mtime)
                    total_mb += (mb := stat.st_size / 1e6)
                    line = f"{name}    {stamp}    {mb:.2f}MB"
                    print(line)
                print(f"Total: {total_mb:.2f}MB".rjust(len(line)))

    def get_options(self):
        return {
            "clear": Clear(),
        }

    def add_parser_arguments(self, parser):
        pass



def on_main(prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


def on_main_until(until, prep=None, ret=None):
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            global _procpass
            r = None
            self.comm.barrier()
            if self.comm.get_rank() == 0:
                r = f(self, *args, **kwargs)
            elif prep:
                prep(self, *args, **kwargs)
            self.comm.barrier()
            while not until(self, *args, **kwargs):
                pass
            if not ret:
                return self.comm.bcast(r, root=0)
            else:
                return ret(self, *args, **kwargs)

        return wrapper

    return decorator


__all__ = ["on_main", "on_main_until"]



class Context:
    def __init__(self, options):
        self.options = options

    def __str__(self):
        opt = "; ".join(f"{opt.name}: {opt.get()}" for opt in self.options.values())
        return f"<Context options({opt})>"


class CLIContext(Context):
    def set_cli_namespace(self, namespace):
        for option in self.options.values():
            for tag in _tags_to_namespace(option.__class__.cli.tags):
                if hasattr(namespace, tag):
                    option.cli = getattr(namespace, tag)
        self.arguments = namespace

    def __getattr__(self, attr):
        for option in self.options.values():
            if option.name == attr:
                return option.get()
        return super().__getattribute__(attr)

    def __str__(self):
        base = super().__str__()
        if not hasattr(self, "arguments"):
            return base[:-1] + " without CLI arguments>"
        return base[:-1] + f" {self.arguments}>"


def reset_cli_context():
    from .options import get_option_descriptors

    for opt in get_option_descriptors().values():
        del opt.cli


def get_cli_context():
    from .options import get_option_descriptors

    return CLIContext(get_option_descriptors())


def _tags_to_namespace(tags):
    yield from (tag.replace("-", "_") for tag in tags)



import abc
import builtins
import inspect
import math
from weakref import WeakKeyDictionary

import numpy as np

from ..exceptions import (
    CastError,
    ClassMapMissingError,
    InvalidReferenceError,
    NoneReferenceError,
    RequirementError,
    TypeHandlingError,
)
from ._compile import _reserved_kw_passes, _wrap_reserved
from ._make import _load_object


class TypeHandler(abc.ABC):
    """
    Base class for any type handler that cannot be described as a single function.

    Declare the `__call__(self, value)` method to convert the given value to the
    desired type, raising a `TypeError` if it failed in an expected manner.

    Declare the `__name__(self)` method to return a name for the type handler to
    display in messages to the user such as errors.

    Declare the optional `__inv__` method to invert the given value back to its
    original value, the type of the original value will usually be lost but the type
    of the returned value can still serve as a suggestion.
    """

    @abc.abstractmethod
    def __call__(self, value):  # pragma: nocover
        pass

    @property
    @abc.abstractmethod
    def __name__(self):  # pragma: nocover
        return "unknown type handler"

    def __inv__(self, value):  # pragma: nocover
        return value

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        call = cls.__call__
        passes = _reserved_kw_passes(call)
        if not all(passes.values()):
            cls.__call__ = _wrap_reserved(call)


def any_():
    def type_handler(value):
        return value

    type_handler.__name__ = "any"
    return type_handler


def in_(container):
    """
    Type validator. Checks whether the given value occurs in the given container.
    Uses the `in` operator.

    :param container: List of possible values
    :type container: list
    :returns: Type validator function
    :rtype: Callable
    """
    error_msg = "a value in: " + builtins.str(container)

    def type_handler(value):
        if value in container:
            return value
        else:
            raise TypeError(f"Couldn't cast '{value}' to " + error_msg)

    type_handler.__name__ = error_msg
    return type_handler


def or_(*type_args):
    """
    Type validator. Attempts to cast the value to any of the given types in order.

    :param type_args: Another type validator
    :type type_args: Callable
    :returns: Type validator function
    :raises: TypeError if none of the given type validators can cast the value.
    :rtype: Callable
    """
    handler_name = "any of: " + ", ".join(x.__name__ for x in type_args)
    # Make sure to wrap all type handlers so that they accept the parent and key args.
    type_args = [_wrap_reserved(t) for t in type_args]

    def type_handler(value, _parent=None, _key=None):
        type_errors = {}
        for t in type_args:
            try:
                v = t(value, _parent=_parent, _key=_key)
            except Exception as e:
                type_error = (
                    builtins.str(e.__class__.__module__)
                    + "."
                    + builtins.str(e.__class__.__name__)
                    + ": "
                    + builtins.str(e)
                )
                type_errors[t.__name__] = type_error
            else:
                return v
        type_errors = "\n".join(
            "- Casting to '{}' raised:\n{}".format(n, e) for n, e in type_errors.items()
        )
        # Use a CastError instead of a TypeError so that the message is passed along as is
        # by upstream error handlers.
        raise CastError(
            "Couldn't cast {} into {}.\n{}".format(value, handler_name, type_errors)
        )

    type_handler.__name__ = handler_name
    return type_handler


class object_(TypeHandler):
    """
    Type validator. Attempts to import the value, absolute, or relative to the
    `module_path` entries.

    :param module_path: List of the modules that should be searched when doing a
      relative import.
    :type module_path: list[str]
    :raises: TypeError when value can't be cast.
    :returns: Type validator function
    :rtype: Callable
    """

    def __init__(self, module_path=None):
        self._module_path = module_path

    def __call__(self, value):
        msg = f"Could not import '{value}': "
        try:
            if isinstance(value, builtins.str):
                obj = _load_object(value, self._module_path)
                obj._cfg_inv = value
            else:
                obj = value
        except Exception as e:
            raise TypeError(msg + builtins.str(e))
        return obj

    def __inv__(self, value):
        return getattr(value, "_cfg_inv", value)

    @property
    def __name__(self):
        return "object"


class class_(object_):
    """
    Type validator. Attempts to import the value as the name of a class, relative to
    the `module_path` entries, absolute or just returning it if it is already a class.

    :param module_path: List of the modules that should be searched when doing a
      relative import.
    :type module_path: list[str]
    :raises: TypeError when value can't be cast.
    :returns: Type validator function
    :rtype: Callable
    """

    def __call__(self, value):
        if inspect.isclass(value):
            return value
        obj = super().__call__(value)
        if not inspect.isclass(obj):
            raise TypeError(f"'{value}' is not a class, got {builtins.type(obj)} instead")
        return obj

    def __inv__(self, value):
        if not inspect.isclass(value):
            value = type(value)
        return f"{value.__module__}.{value.__name__}"

    @property
    def __name__(self):
        return "class"


class function_(object_):
    """
    Type validator. Attempts to import the value, absolute, or relative to the
    `module_path` entries, and verifies that it is callable.

    :param module_path: List of the modules that should be searched when doing a
      relative import.
    :type module_path: list[str]
    :raises: TypeError when value can't be cast.
    :returns: Type validator function
    :rtype: Callable
    """

    def __call__(self, value):
        if callable(value):
            return value
        obj = super().__call__(value)
        if not callable(obj):
            raise TypeError(f"Could not import {value} as a callable function.")
        return obj

    def __inv__(self, value):
        return f"{value.__module__}.{value.__name__}"

    @property
    def __name__(self):
        return "function"


class method(function_):
    def __init__(self, class_name):
        super().__init__()
        self._class = class_name

    def __call__(self, value):
        parent = class_()(self._class)
        try:
            obj = getattr(parent, value)
        except AttributeError as e:
            raise TypeError(builtins.str(e)) from None
        if not callable(obj):
            raise TypeError(f"Could not import '{value}' as a method of `{self._class}`.")
        return obj

    def __inv__(self, value):
        return value.__name__

    @property
    def __name__(self):
        return f"method of '{self._class}'"


class WeakInverter:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._map = WeakKeyDictionary()

    def store_value(self, value, result):
        self._map[result] = value

    def __inv__(self, value):
        return self._map.get(value, value)


class method_shortcut(method, function_):
    def __call__(self, value):
        try:
            obj = method.__call__(self, value)
        except TypeError:
            try:
                obj = function_.__call__(self, value)
            except TypeError:
                raise TypeError(
                    f"Could not import '{value}' as a function or a method of `{self._class}`."
                ) from None
        return obj

    def __inv__(self, value):
        if inspect.isfunction(value):
            try:
                method.__call__(self, value.__name__)
                return method.__inv__(self, value)
            except TypeError:
                return function_.__inv__(self, value)
        else:
            return value


def str(strip=False, lower=False, upper=False, safe=True):
    """
    Type validator. Attempts to cast the value to a str, optionally with some sanitation.

    :param strip: Trim whitespaces
    :type strip: bool
    :param lower: Convert value to lowercase
    :type lower: bool
    :param upper: Convert value to uppercase
    :type upper: bool
    :param safe: If False, checks that the type of value is string before cast.
    :type safe: bool
    :returns: Type validator function
    :raises: TypeError when value can't be cast.
    :rtype: Callable
    """
    handler_name = "str"
    # Compile a custom function to sanitize the string according to args
    safety_check = "\n if not isinstance(s, str):\n  raise TypeError()\n" if safe else ""
    fstr = f"def f(s):{safety_check} return str(s)"
    for add, mod in zip((strip, lower, upper), ("strip", "lower", "upper")):
        if add:
            fstr += f".{mod}()"
    go_fish = builtins.dict()
    exec(compile(fstr, "__", "exec"), go_fish)
    f = go_fish["f"]

    def type_handler(value):
        return f(value)

    type_handler.__name__ = handler_name
    return type_handler


def int(min=None, max=None):
    """
    Type validator. Attempts to cast the value to an int, optionally within some
    bounds.

    :param min: Minimum valid value
    :type min: int
    :param max: Maximum valid value
    :type max: int
    :returns: Type validator function
    :raises: TypeError when value can't be cast.
    :rtype: Callable
    """
    handler_name = "int"
    if min is not None and max is not None:
        handler_name += " between [{}, {}]".format(min, max)
    elif min is not None:
        handler_name += " >= {}".format(min)
    elif max is not None:
        handler_name += " <= {}".format(max)

    def type_handler(value):
        try:
            v = builtins.int(value)
            if min is not None and min > v or max is not None and max < v:
                raise Exception()
            return v
        except Exception:
            raise TypeError(
                "Could not cast {} to an {}.".format(value, handler_name)
            ) from None

    type_handler.__name__ = handler_name
    return type_handler


_float = float


def float(min=None, max=None):
    """
    Type validator. Attempts to cast the value to an float, optionally within some
    bounds.

    :param min: Minimum valid value
    :type min: float
    :param max: Maximum valid value
    :type max: float
    :returns: Type validator function
    :raises: TypeError when value can't be cast.
    :rtype: Callable
    """
    handler_name = "float"
    if min is not None and max is not None:
        handler_name += " between [{}, {}]".format(min, max)
    elif min is not None:
        handler_name += " >= {}".format(min)
    elif max is not None:
        handler_name += " <= {}".format(max)

    def type_handler(value):
        try:
            v = _float(value)
            if min is not None and min > v or max is not None and max < v:
                raise Exception()
            return v
        except Exception:
            raise TypeError("Could not cast {} to an {}.".format(value, handler_name))

    type_handler.__name__ = handler_name
    return type_handler


def number(min=None, max=None):
    """
    Type validator. If the given value is an int returns an int, tries to cast to float
    otherwise

    :param min: Minimum valid value
    :type min: float
    :param max: Maximum valid value
    :type max: float
    :returns: Type validator function
    :raises: TypeError when value can't be cast.
    :rtype: Callable
    """
    handler_name = "number"
    if min is not None and max is not None:
        handler_name += " between [{}, {}]".format(min, max)
    elif min is not None:
        handler_name += " >= {}".format(min)
    elif max is not None:
        handler_name += " <= {}".format(max)

    def type_handler(value):
        try:
            if isinstance(value, builtins.int):
                v = builtins.int(value)
                if min is not None and min > v or max is not None and max < v:
                    raise Exception()
            else:
                v = _float(value)
                if min is not None and min > v or max is not None and max < v:
                    raise Exception()
            return v
        except Exception:
            raise TypeError("Could not cast {} to a {}.".format(value, handler_name))

    type_handler.__name__ = handler_name
    return type_handler


def key():
    """
    Type handler for keys in configuration trees. Keys can be either int indices of a
    config list, or string keys of a config dict.

    :returns: Type validator function
    :rtype: Callable
    """

    def type_handler(value):
        if not (isinstance(value, builtins.int) or isinstance(value, builtins.str)):
            raise TypeError(f"{type(value)} is not an int or str")
        else:
            return value

    type_handler.__name__ = "configuration key"
    return type_handler


def scalar_expand(scalar_type, size=None, expand=None):
    """
    Create a method that expands a scalar into an array with a specific size or uses
    an expansion function.

    :param scalar_type: Type of the scalar
    :type scalar_type: type
    :param size: Expand the scalar to an array of a fixed size.
    :type size: int
    :param expand: A function that takes the scalar value as argument and returns the
      expanded form.
    :type expand: Callable
    :returns: Type validator function
    :rtype: Callable
    """

    if expand is None:

        def expand(x):
            return [1.0] * x

    def type_handler(value):
        # No try block: let it raise the cast error.
        v = scalar_type(value)
        # Expand the scalar.
        return expand(v)

    type_handler.__name__ = "expanded list of " + scalar_type.__name__
    return type_handler


def list_or_scalar(scalar_type, size=None):
    """
    Type validator that accepts a scalar or list of said scalars.

    :param scalar_type: Type of the scalar
    :type scalar_type: type
    :param size: Expand the scalar to an array of a fixed size.
    :type size: int
    :returns: Type validator function
    :rtype: Callable
    """
    type_handler = or_(list(scalar_type, size), scalar_type)

    type_handler.__name__ += " or " + scalar_type.__name__
    return type_handler


def voxel_size():
    return list_or_scalar(float(), 3)


def list(type=builtins.str, size=None):
    """
    Type validator for lists. Type casts each element to the given type and optionally
    validates the length of the list.

    :param type: Type validator of the elements.
    :type type: Callable
    :param size: Mandatory length of the list.
    :type size: int
    :returns: Type validator function
    :rtype: Callable
    """

    def type_handler(value):
        # Simple lists default to returning None for None, while configuration lists
        # default to an empty list.
        if value is None:
            return None
        v = builtins.list(value)
        try:
            for i, e in enumerate(v):
                v[i] = type(e)
        except Exception:
            raise TypeError(
                "Couldn't cast element {} of {} into {}".format(i, value, type.__name__)
            )
        if size is not None and len(v) != size:
            raise ValueError(
                "Couldn't cast {} into a {} element list".format(value, size)
            )
        return v

    type_handler.__name__ = "list{} of {}".format(
        "[{}]".format(size) if size is not None else "", type.__name__
    )
    return type_handler


def dict(type=builtins.str):
    """
    Type validator for dicts. Type casts each element to the given type.

    :param type: Type validator of the elements.
    :type type: Callable
    :returns: Type validator function
    :rtype: Callable
    """

    def type_handler(value):
        if value is None:
            return None
        v = builtins.dict(value)
        try:
            for k, e in v.items():
                v[k] = type(e)
        except Exception:
            raise TypeError(
                "Couldn't cast {} of {} into {}".format(k, value, type.__name__)
            )
        return v

    type_handler.__name__ = "dict of {}".format(type.__name__)
    return type_handler


def fraction():
    """
    Type validator. Type casts the value into a rational number between 0 and 1
    (inclusive).

    :returns: Type validator function
    :rtype: Callable
    """

    def type_handler(value):
        v = _float(value)
        if v < 0.0 or v > 1.0:
            raise ValueError("{} is out of the 0-1 range for a fraction.".format(value))
        return v

    type_handler.__name__ = "fraction [0.; 1.]"
    return type_handler


class deg_to_radian(TypeHandler):
    """
    Type validator. Type casts the value from degrees to radians.
    """

    def __call__(self, value):
        v = _float(value)
        return v * 2 * math.pi / 360

    @property
    def __name__(self):  # pragma: nocover
        return "degrees"

    def __inv__(self, value):
        v = _float(value)
        return v * 360 / (2 * math.pi)


class distribution(TypeHandler):
    """
    Type validator. Type casts the value or node to a distribution.
    """

    def __call__(self, value, _key=None, _parent=None):
        from ._distributions import Distribution

        if not isinstance(value, builtins.list) and not isinstance(value, builtins.dict):
            value = {"distribution": "constant", "constant": value}

        return Distribution(**value, _key=_key, _parent=_parent)

    @property
    def __name__(self):  # pragma: nocover
        return "distribution"

    def __inv__(self, value):
        if value["distribution"] == "constant":
            return value["constant"]
        else:
            return value


class evaluation(TypeHandler):
    """
    Type validator. Provides a structured way to evaluate a python statement from the
    config. The evaluation context provides ``numpy`` as ``np``.

    :returns: Type validator function
    :rtype: Callable
    """

    def __init__(self):
        self._references = {}

    def __call__(self, value):
        cfg = builtins.dict(value)
        statement = builtins.str(cfg.get("statement", "None"))
        locals_ = builtins.dict(cfg.get("variables", {}))
        globals_ = {"np": np}
        res = eval(statement, globals_, locals_)
        self._references[id(res)] = value
        return res

    @property
    def __name__(self):
        return "evaluation"

    def get_original(self, value):
        """
        Return the original configuration node associated with the given evaluated value.

        :param value: A value that was produced by this type handler.
        :type value: Any
        :raises: NoneReferenceError when `value` is `None`, InvalidReferenceError when
          there is no config associated to the object id of this value.
        """
        # None is a singleton, so it's not bijective, it's also the value returned when
        # a weak reference is removed; so it's doubly unsafe to check for references to it
        if value is None:
            raise NoneReferenceError("Can't create bijection for NoneType value.")
        vid = id(value)
        # Create a set of references from our stored weak references that are still alive.
        if vid not in self._references:
            raise InvalidReferenceError(f"No evaluation reference found for {vid}", value)
        return self._references[vid]

    def __inv__(self, value):
        try:
            return self.get_original(value)
        except TypeHandlingError:
            # Original does not exist or can't be obtained, just return the given value.
            return value


def in_classmap():
    """
    Type validator. Checks whether the given string occurs in the class map of a
    dynamic node.

    :returns: Type validator function
    :rtype: Callable
    """

    def type_handler(value, _parent, _key=None):
        class_name = _parent.__class__.__name__
        if not hasattr(_parent.__class__, "_config_dynamic_classmap"):
            raise ClassMapMissingError(
                f"Class map missing for `{class_name}`,"
                + " required when using `in_classmap` type handler."
            )
        classmap = _parent.__class__._config_dynamic_classmap
        if value not in classmap:
            classmap_str = ", ".join(f"'{key}'" for key in classmap)
            raise CastError(
                f"'{value}' is not a valid classmap identifier for `{class_name}`."
                + f" Choose from: {classmap_str}"
            )
        return value

    type_handler.__name__ = "a classmap value"
    return type_handler


def mut_excl(*mutuals, required=True, max=1, shortform=False):
    """
    Requirement handler for mutually exclusive attributes.

    :param str mutuals: The keys of the mutually exclusive attributes.
    :param bool required: Whether at least one of the keys is required
    :param int max: The maximum amount of keys that may occur together.
    :param bool shortform: Allow the short form alternative.
    :returns: Requirement function
    :rtype: Callable
    """
    listed = ", ".join(f"`{m}`" for m in mutuals[:-1])
    if len(mutuals) > 1:
        listed += f" {{}} `{mutuals[-1]}`"

    def requirement(section):
        if shortform and section.is_shortform:
            return False
        bools = [m in section for m in mutuals]
        given = sum(bools)
        if given > max:
            if max > 1:
                err_msg = f"Maximum {max} of {listed} may be specified. {given} given."
            else:
                err_msg = f"The {listed} attributes are mutually exclusive."
            err_msg = err_msg.format("and")
            raise RequirementError(err_msg)
        if not given and required:
            err_msg = f"A {listed} attribute is required."
            raise RequirementError(err_msg)
        return False

    return requirement


def shortform():
    def requirement(section):
        return not section.is_shortform

    return requirement


class ndarray(TypeHandler):
    """
    Type validator numpy arrays.

    :returns: Type validator function
    :rtype: Callable
    """

    def __call__(self, value):
        return np.array(value, copy=False)

    @property
    def __name__(self):
        return "ndarray"

    def __inv__(self, value):
        return value.tolist()


def none():
    def type_handler(value, _parent, _key=None):
        if value is not None:
            raise TypeError("value is not None")
        return value

    type_handler.__name__ = "a None value"
    return type_handler


class PackageRequirement(TypeHandler):
    def __call__(self, value):
        from packaging.requirements import Requirement

        return Requirement(value)

    @property
    def __name__(self):
        return "package requirement"

    def __inv__(self, value):
        return str(value)

    def __hint__(self):
        return "numpy==1.24.0"


__all__ = [
    "PackageRequirement",
    "TypeHandler",
    "WeakInverter",
    "any_",
    "class_",
    "deg_to_radian",
    "dict",
    "distribution",
    "evaluation",
    "float",
    "fraction",
    "function_",
    "in_",
    "in_classmap",
    "int",
    "key",
    "list",
    "list_or_scalar",
    "method",
    "method_shortcut",
    "mut_excl",
    "ndarray",
    "none",
    "number",
    "object_",
    "or_",
    "scalar_expand",
    "shortform",
    "str",
    "voxel_size",
]
__api__ = ["PackageRequirement", "TypeHandler", "WeakInverter"]



import functools
import math
import typing

import numpy as np
from numpy.random import default_rng

from .. import config
from ..config import refs, types

if typing.TYPE_CHECKING:
    from .cell import CellModel


@config.dynamic(attr_name="strategy", default="all", auto_classmap=True)
class Targetting:
    type: typing.Union[typing.Literal["cell"], typing.Literal["connection"]] = (
        config.attr(type=types.in_(["cell", "connection"]), default="cell")
    )

    def get_targets(self, adapter, simulation, simdata):
        if self.type == "cell":
            return simdata.populations
        elif self.type == "connection":
            return simdata.connections


@config.node
class CellTargetting(Targetting, classmap_entry="all"):
    @config.property
    def type(self):
        return "cell"

    def get_targets(self, adapter, simulation, simdata):
        return simdata.populations


@config.node
class ConnectionTargetting(Targetting, classmap_entry="all_connections"):
    @config.property
    def type(self):
        return "connection"

    def get_targets(self, adapter, simulation, simdata):
        return simdata.connections


class CellModelFilter:
    cell_models: list["CellModel"] = config.reflist(
        refs.sim_cell_model_ref, required=False
    )

    def get_targets(self, adapter, simulation, simdata):
        return {
            model: pop
            for model, pop in simdata.populations.items()
            if not self.cell_models or model in self.cell_models
        }


class FractionFilter:
    count = config.attr(
        type=int, required=types.mut_excl("fraction", "count", required=False)
    )
    fraction = config.attr(
        type=types.fraction(),
        required=types.mut_excl("fraction", "count", required=False),
    )

    def satisfy_fractions(self, targets):
        return {model: self._frac(data) for model, data in targets.items()}

    def _frac(self, data):
        take = None
        if self.count is not None:
            take = self.count
        if self.fraction is not None:
            take = math.floor(len(data) * self.fraction)
        if take is None:
            return data
        else:
            # Select `take` elements from data with a boolean mask (otherwise a sorted
            # integer mask would be required)
            idx = np.zeros(len(data), dtype=bool)
            idx[np.random.default_rng().integers(0, len(data), take)] = True
            return data[idx]

    @staticmethod
    def filter(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            return self.satisfy_fractions(f(self, *args, **kwargs))

        return wrapper


@config.node
class CellModelTargetting(
    CellModelFilter, FractionFilter, CellTargetting, classmap_entry="cell_model"
):
    """
    Targets all cells of certain cell models.
    """

    cell_models: list["CellModel"] = config.reflist(
        refs.sim_cell_model_ref, required=True
    )

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        return super().get_targets(adapter, simulation, simdata)


@config.node
class RepresentativesTargetting(
    CellModelFilter, FractionFilter, CellTargetting, classmap_entry="representatives"
):
    """
    Targets all identifiers of certain cell types.
    """

    n: int = config.attr(type=int, default=1)

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        return {
            model: default_rng().choice(len(pop), size=self.n, replace=False)
            for model, pop in super().get_targets(adapter, simulation, simdata)
        }


@config.node
class ByIdTargetting(FractionFilter, CellTargetting, classmap_entry="by_id"):
    """
    Targets all given identifiers.
    """

    ids: dict[str, list[int]] = config.attr(
        type=types.dict(type=types.list(type=int)), required=True
    )

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        by_name = {model.name: model for model in simdata.populations.keys()}
        return {
            model: simdata.populations[model][ids]
            for model_name, ids in self.ids.items()
            if (model := by_name.get(model_name)) is not None
        }


@config.node
class ByLabelTargetting(
    CellModelFilter, FractionFilter, CellTargetting, classmap_entry="by_label"
):
    """
    Targets all given labels.
    """

    labels: list[str] = config.attr(type=types.list(type=str), required=True)

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        return {
            model: simdata.populations[
                simdata.placement[model].get_label_mask(self.labels)
            ]
            for model in super().get_targets(adapter, simulation, simdata).keys()
        }


@config.node
class CylindricalTargetting(
    CellModelFilter, FractionFilter, CellTargetting, classmap_entry="cylinder"
):
    """
    Targets all cells in a cylinder along specified axis.
    """

    origin: list[float] = config.attr(type=types.list(type=float, size=2))
    axis: typing.Union[typing.Literal["x"], typing.Literal["y"], typing.Literal["z"]] = (
        config.attr(type=types.in_(["x", "y", "z"]), default="y")
    )
    radius: float = config.attr(type=float, required=True)

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        """
        Target all or certain cells within a cylinder of specified radius.
        """
        if self.axis == "x":
            axes = [1, 2]
        elif self.axis == "y":
            axes = [0, 2]
        else:
            axes = [0, 1]
        return {
            model: simdata.populations[model][
                np.sum(
                    simdata.placement[model].load_positions()[:, axes] - self.origin**2,
                    axis=0,
                )
                < self.radius**2
            ]
            for model in super().get_targets(adapter, simulation, simdata).keys()
        }


@config.node
class SphericalTargetting(
    CellModelFilter, FractionFilter, CellTargetting, classmap_entry="sphere"
):
    """
    Targets all cells in a sphere.
    """

    origin: list[float] = config.attr(type=types.list(type=float, size=3), required=True)
    radius: float = config.attr(type=float, required=True)

    @FractionFilter.filter
    def get_targets(self, adapter, simulation, simdata):
        """
        Target all or certain cells within a sphere of specified radius.
        """
        return {
            model: simdata.populations[model][
                (
                    np.sum(
                        (simdata.placement[model].load_positions() - self.origin) ** 2,
                        axis=1,
                    )
                    < self.radius**2
                )
            ]
            for model in super().get_targets(adapter, simulation, simdata).keys()
        }


@config.dynamic(
    attr_name="strategy",
    default="everywhere",
    auto_classmap=True,
    classmap_entry="everywhere",
)
class LocationTargetting:
    def get_locations(self, cell):
        return cell.locations


@config.node
class SomaTargetting(LocationTargetting, classmap_entry="soma"):
    def get_locations(self, cell):
        return [cell.locations[(0, 0)]]


@config.node
class LabelTargetting(LocationTargetting, classmap_entry="label"):
    labels = config.list(required=True)

    def get_locations(self, cell):
        locs = [
            loc
            for loc in cell.locations.values()
            if all(l in loc.section.labels for l in self.labels)
        ]
        return locs


@config.node
class BranchLocTargetting(LabelTargetting, classmap_entry="branch"):
    x = config.attr(type=types.fraction(), default=0.5)

    def get_locations(self, cell):
        locations = super().get_locations(cell)
        branches = set()
        selected = []
        for loc in locations:
            if (
                loc._loc[0] not in branches
                and loc.arc(0) <= self.x
                and loc.arc(1) > self.x
            ):
                selected.append(loc)
                branches.add(loc._loc[0])
        return selected


__all__ = [
    "BranchLocTargetting",
    "ByIdTargetting",
    "ByLabelTargetting",
    "CellModelFilter",
    "CellModelTargetting",
    "CellTargetting",
    "ConnectionTargetting",
    "CylindricalTargetting",
    "FractionFilter",
    "LabelTargetting",
    "LocationTargetting",
    "RepresentativesTargetting",
    "SomaTargetting",
    "SphericalTargetting",
    "Targetting",
]



import abc
import functools
import typing
from pathlib import Path

import numpy as np

from .. import config, plugins
from .._util import immutable, obj_str_insert
from ..trees import BoxTree
from ._chunks import Chunk

if typing.TYPE_CHECKING:
    from ..cell_types import CellType


@config.pluggable(key="engine", plugin_name="storage engine")
class StorageNode:
    root: typing.Any = config.slot()

    @classmethod
    def __plugins__(cls):
        if not hasattr(cls, "_plugins"):
            cls._plugins = {
                name: plugin.StorageNode
                for name, plugin in plugins.discover("storage.engines").items()
            }
        return cls._plugins


class Interface(abc.ABC):
    _iface_engine_key = None

    def __init__(self, engine):
        self._engine = engine

    def __init_subclass__(cls, **kwargs):
        # Only change engine key if explicitly given.
        if "engine_key" in kwargs:
            cls._iface_engine_key = kwargs["engine_key"]


class NoopLock:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Engine(Interface):
    """
    Engines perform the transactions that come from the storage object, and read/write
    data in a specific format. They can perform collective or individual actions.

    .. warning::

      Collective actions can only be performed from all nodes, or deadlocks occur. This
      means in particular that they may not be called from component code.

    """

    def __init__(self, root, comm):
        self._root = comm.bcast(root, root=0)
        self._comm = comm
        self._readonly = False

    def __eq__(self, other):
        eq_format = self._format == getattr(other, "_format", None)
        eq_root = self._root == getattr(other, "_root", None)
        return eq_format and eq_root

    @property
    def root(self):
        """
        The unique identifier for the storage. Usually pathlike, but can be anything.
        """
        return self._root

    @property
    def comm(self):
        """
        The communicator in charge of collective operations.
        """
        return self._comm

    def set_comm(self, comm):
        """
        :guilabel:`collective` Set a new communicator in charge of collective operations.
        """
        self._comm = comm

    @property
    def format(self):
        """
        Name of the type of engine. Automatically set through the plugin system.
        """
        return self._format

    @property
    @abc.abstractmethod
    def root_slug(self):
        """
        Must return a pathlike unique identifier for the root of the storage object.
        """
        pass

    @abc.abstractmethod
    def recognizes(self, root):
        """
        Must return whether the given argument is recognized as a valid storage object.
        """
        pass

    @classmethod
    def peek_exists(cls, root):
        """
        Must peek at the existence of the given root, without instantiating anything.
        """
        try:
            return Path(root).exists()
        except Exception:
            return False

    @abc.abstractmethod
    def exists(self):
        """
        Must check existence of the storage object.
        """
        pass

    @abc.abstractmethod
    def create(self):
        """
        :guilabel:`collective` Must create the storage engine.
        """
        pass

    @abc.abstractmethod
    def move(self, new_root):
        """
        :guilabel:`collective` Must move the storage object to the new root.
        """
        pass

    @abc.abstractmethod
    def copy(self, new_root):
        """
        :guilabel:`collective` Must copy the storage object to the new root.
        """
        pass

    @abc.abstractmethod
    def remove(self):
        """
        :guilabel:`collective` Must remove the storage object.
        """
        pass

    @abc.abstractmethod
    def clear_placement(self):
        """
        :guilabel:`collective` Must clear existing placement data.
        """
        pass

    @abc.abstractmethod
    def clear_connectivity(self):
        """
        :guilabel:`collective` Must clear existing connectivity data.
        """
        pass

    @abc.abstractmethod
    def get_chunk_stats(self):
        """
        :guilabel:`readonly` Must return a dictionary with all chunk statistics.
        """
        pass

    def read_only(self):
        """
        A context manager that enters the engine into readonly mode. In
        readonly mode the engine does not perform any locking, write-operations or network
        synchronization, and errors out if a write operation is attempted.
        """
        self._readonly = True
        return ReadOnlyManager(self)

    def readwrite(self):
        self._readonly = False


class ReadOnlyManager:
    def __init__(self, engine):
        self._e = engine

    def __enter__(self):
        self._e._readonly = True

    def __exit__(self, *args):
        self._e._readonly = False


class NetworkDescription(Interface):
    pass


class FileStore(Interface, engine_key="files"):
    """
    Interface for the storage and retrieval of files essential to the network description.
    """

    @abc.abstractmethod
    def all(self):
        """
        Return all ids and associated metadata in the file store.
        """
        pass

    @abc.abstractmethod
    def store(self, content, id=None, meta=None, encoding=None, overwrite=False):
        """
        Store content in the file store. Should also store the current timestamp as
        `mtime` meta.

        :param content: Content to be stored
        :type content: str
        :param id: Optional specific id for the content to be stored under.
        :type id: str
        :param meta: Metadata for the content
        :type meta: dict
        :param encoding: Optional encoding
        :type encoding: str
        :param overwrite: Overwrite existing file
        :type overwrite: bool
        :returns: The id the content was stored under
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def load(self, id):
        """
        Load the content of an object in the file store.

        :param id: id of the content to be loaded.
        :type id: str
        :returns: The content of the stored object
        :rtype: str
        :raises FileNotFoundError: The given id doesn't exist in the file store.
        """
        pass

    @abc.abstractmethod
    def remove(self, id):
        """
        Remove the content of an object in the file store.

        :param id: id of the content to be removed.
        :type id: str
        :raises FileNotFoundError: The given id doesn't exist in the file store.
        """
        pass

    @abc.abstractmethod
    def store_active_config(self, config):
        """
        Store configuration in the file store and mark it as the active configuration of
        the stored network.

        :param config: Configuration to be stored
        :type config: :class:`~.config.Configuration`
        :returns: The id the config was stored under
        :rtype: str
        """
        pass

    @abc.abstractmethod
    def load_active_config(self):
        """
        Load the active configuration stored in the file store.

        :returns: The active configuration
        :rtype: :class:`~.config.Configuration`
        :raises Exception: When there's no active configuration in the file store.
        """
        pass

    @abc.abstractmethod
    def has(self, id):
        """
        Must return whether the file store has a file with the given id.
        """
        pass

    @abc.abstractmethod
    def get_mtime(self, id):
        """
        Must return the last modified timestamp of file with the given id.
        """
        pass

    @abc.abstractmethod
    def get_encoding(self, id):
        """
        Must return the encoding of the file with the given id, or None if it is
        unspecified binary data.
        """
        pass

    @abc.abstractmethod
    def get_meta(self, id) -> typing.Mapping[str, typing.Any]:
        """
        Must return the metadata of the given id.
        """
        pass

    def get(self, id) -> "StoredFile":
        """
        Return a StoredFile wrapper
        """
        if not self.has(id):
            raise FileNotFoundError(f"File with id '{id}' not found.")
        return StoredFile(self, id)

    def find_files(self, predicate):
        return (
            StoredFile(self, id_) for id_, m in self.all().items() if predicate(id_, m)
        )

    def find_file(self, predicate):
        return next(self.find_files(predicate), None)

    def find_id(self, id):
        return self.find_file(lambda id_, _: id_ == id)

    def find_meta(self, key, value):
        return self.find_file(lambda _, meta: meta.get(key, None) == value)


class StoredFile:
    def __init__(self, store, id):
        self.store = store
        self.id = id

    @property
    def meta(self):
        return self.store.get_meta(self.id)

    @property
    def mtime(self):
        return self.store.get_mtime(self.id)

    def load(self):
        return self.store.load(self.id)


class PlacementSet(Interface):
    """
    Interface for the storage of placement data of a cell type.
    """

    @abc.abstractmethod
    def __init__(self, engine, cell_type):
        self._engine = engine
        self._type = cell_type
        self._tag = cell_type.name

    @abc.abstractmethod
    def __len__(self):
        pass

    @obj_str_insert
    def __repr__(self):
        cell_type = self.cell_type
        try:
            ms = self.load_morphologies()
        except Exception:
            return f"cell type: '{cell_type.name}'"
        if not len(ms):
            mstr = "without morphologies"
        else:
            mstr = f"with {len(ms._loaders)} morphologies"
        return f"cell type: '{cell_type.name}', {len(self)} cells, {mstr}"

    @property
    def cell_type(self):
        """
        The associated cell type.

        :returns: The cell type
        :rtype: ~bsb.cell_types.CellType
        """
        return self._type

    @property
    def tag(self):
        """
        The unique identifier of the placement set.

        :returns: Unique identifier
        :rtype: str
        """
        return self._tag

    @classmethod
    @abc.abstractmethod
    def create(cls, engine, cell_type):
        """
        Create a placement set.

        :param engine: The engine that governs this PlacementSet.
        :type engine: `bsb.storage.interfaces.Engine`
        :param cell_type: The cell type whose data is stored in the placement set.
        :type cell_type: bsb.cell_types.CellType
        :returns: A placement set
        :rtype: bsb.storage.interfaces.PlacementSet
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def exists(engine, cell_type):
        """
        Check existence of a placement set.

        :param engine: The engine that governs the existence check.
        :type engine: `bsb.storage.interfaces.Engine`
        :param cell_type: The cell type to look for.
        :type cell_type: bsb.cell_types.CellType
        :returns: Whether the placement set exists.
        :rtype: bool
        """
        pass

    @classmethod
    def require(cls, engine, type):
        """
        Return and create a placement set, if it didn't exist before.

        The default implementation uses the
        :meth:`~bsb.storage.interfaces.PlacementSet.exists` and
        :meth:`~bsb.storage.interfaces.PlacementSet.create` methods.

        :param engine: The engine that governs this PlacementSet.
        :type engine: `bsb.storage.interfaces.Engine`
        :param cell_type: The cell type whose data is stored in the placement set.
        :type cell_type: bsb.cell_types.CellType
        :returns: A placement set
        :rtype: bsb.storage.interfaces.PlacementSet
        """
        if not cls.exists(engine, type):
            cls.create(engine, type)
        return cls(engine, type)

    @abc.abstractmethod
    def clear(self, chunks=None):
        """
        Clear (some chunks of) the placement set.

        :param chunks: If given, the specific chunks to clear.
        :type chunks: List[bsb.storage._chunks.Chunk]
        """
        pass

    @abc.abstractmethod
    def get_all_chunks(self):
        """
        Get all the chunks that exist in the placement set.

        :returns: List of existing chunks.
        :rtype: List[bsb.storage._chunks.Chunk]
        """
        pass

    @abc.abstractmethod
    def load_ids(self):
        pass

    @abc.abstractmethod
    def load_positions(self):
        """
        Return a dataset of cell positions.

        :returns: An (Nx3) dataset of positions.
        :rtype: numpy.ndarray
        """
        pass

    @abc.abstractmethod
    def load_rotations(self):
        """
        Load the rotation data of the placement set
        :returns: A rotation set
        :rtype: ~bsb.morphologies.RotationSet
        """
        pass

    @abc.abstractmethod
    def load_morphologies(self, allow_empty=False):
        """
        Return a :class:`~.morphologies.MorphologySet` associated to the cells. Raises an
        error if there is no morphology data, unless `allow_empty=True`.

        :param bool allow_empty: Silence missing morphology data error, and return an
          empty morphology set.
        :returns: Set of morphologies
        :rtype: :class:`~.morphologies.MorphologySet`
        """
        pass

    @abc.abstractmethod
    def load_additional(self, key=None):
        pass

    def count_morphologies(self):
        """
        Must return the number of different morphologies used in the set.
        """
        return self.load_morphologies(allow_empty=True).count_morphologies()

    @abc.abstractmethod
    def __iter__(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def append_data(
        self,
        chunk,
        positions=None,
        morphologies=None,
        rotations=None,
        additional=None,
        count=None,
    ):
        """
        Append data to the placement set. If any of ``positions``, ``morphologies``, or
        ``rotations`` is given, the arguments to its left must also be given (e.g. passing
        morphologies, but no positions, is not allowed, passing just positions is allowed)

        :param chunk: The chunk to store data in.
        :type chunk: ~bsb.storage._chunks.Chunk
        :param positions: Cell positions
        :type positions: numpy.ndarray
        :param rotations: Cell rotations
        :type rotations: ~bsb.morphologies.RotationSet
        :param morphologies: Cell morphologies
        :type morphologies: ~bsb.morphologies.MorphologySet
        :param additional: Additional datasets with 1 value per cell, will be stored
          under its key in the dictionary
        :type additional: Dict[str, numpy.ndarray]
        :param count: Amount of entities to place. Excludes the use of any positional,
          rotational or morphological data.
        :type count: int
        """
        pass

    @abc.abstractmethod
    def append_additional(self, name, chunk, data):
        """
        Append arbitrary user data to the placement set. The length of the data must match
        that of the placement set, and must be storable by the engine.

        :param name:
        :param chunk: The chunk to store data in.
        :type chunk: ~bsb.storage._chunks.Chunk
        :param data: Arbitrary user data. You decide |:heart:|
        :type data: numpy.ndarray
        """
        pass

    @abc.abstractmethod
    def chunk_context(self, chunks):
        pass

    @abc.abstractmethod
    def set_chunk_filter(self, chunks):
        """
        Should limit the scope of the placement set to the given chunks.

        :param chunks: List of chunks
        :type chunks: list[bsb.storage._chunks.Chunk]
        """
        pass

    @abc.abstractmethod
    def set_label_filter(self, labels):
        """
        Should limit the scope of the placement set to the given labels.

        :param labels: List of labels
        :type labels: list[str]
        """
        pass

    @abc.abstractmethod
    def set_morphology_label_filter(self, morphology_labels):
        """
        Should limit the scope of the placement set to the given sub-cellular labels. The
        morphologies returned by
        :meth:`~.storage.interfaces.PlacementSet.load_morphologies` should return a
        filtered form of themselves if :meth:`~.morphologies.Morphology.as_filtered` is
        called on them.

        :param morphology_labels: List of labels
        :type morphology_labels: list[str]
        """
        pass

    @abc.abstractmethod
    def label(self, labels, cells):
        """
        Should label the cells with given labels.

        :param cells: Array of cells in this set to label.
        :type cells: numpy.ndarray
        :param labels: List of labels
        :type labels: list[str]
        """
        pass

    @abc.abstractmethod
    def get_labelled(self, labels):
        """
        Should return the cells labelled with given labels.

        :param cells: Array of cells in this set to label.
        :type cells: numpy.ndarray
        :param labels: List of labels
        :type labels: list[str]
        """
        pass

    @abc.abstractmethod
    def get_label_mask(self, labels):
        """
        Should return a mask that fits the placement set for the cells with given labels.

        :param cells: Array of cells in this set to label.
        :type cells: numpy.ndarray
        :param labels: List of labels
        :type labels: list[str]
        """
        pass

    @abc.abstractmethod
    def get_chunk_stats(self):
        """
        Should return how many cells were placed in each chunk.
        """
        pass

    def load_boxes(self, morpho_cache=None):
        """
        Load the cells as axis aligned bounding box rhomboids matching the extension,
        orientation and position in space. This function loads morphologies, unless a
        `morpho_cache` is given, then that is used.

        :param morpho_cache: If you've previously loaded morphologies with soft or hard
          caching enabled, you can pass the resulting morphology set here to reuse it. If
          afterwards you need the morphology set, you best call :meth:`.load_morphologies`
          first and reuse it here.
        :type morpho_cache: ~bsb.morphologies.MorphologySet
        :returns: An iterator with 6 coordinates per cell: 3 min and 3 max coords, the
          bounding box of that cell's translated and rotated morphology.
        :rtype: Iterator[Tuple[float, float, float, float, float, float]]
        :raises: DatasetNotFoundError if no morphologies are found.
        """
        if morpho_cache is None:
            mset = self.load_morphologies()
        else:
            mset = morpho_cache
        expansion = [*zip([0] * 4 + [1] * 4, ([0] * 2 + [1] * 2) * 2, [0, 1] * 4)]

        def _box_of(m, o, r):
            oo = (m["ldc"], m["mdc"])
            # Make the 8 corners of the box
            corners = np.array([[oo[x][0], oo[y][1], oo[z][2]] for x, y, z in expansion])
            # Rotate them
            rotbox = r.apply(corners)
            # Find outer box, by rotating and translating the starting box
            return np.concatenate(
                (np.min(rotbox, axis=0) + o, np.max(rotbox, axis=0) + o)
            )

        iters = (mset.iter_meta(), self.load_positions(), self.load_rotations())
        return map(_box_of, *iters)

    def load_box_tree(self, morpho_cache=None):
        """
        Load boxes, and form an RTree with them, for fast spatial lookup of rhomboid
        intersection.

        :param morpho_cache: See :meth:`~bsb.storage.interfaces.PlacementSet.load_boxes`.
        :returns: A boxtree
        :rtype: bsb.trees.BoxTree
        """
        return BoxTree(list(self.load_boxes(morpho_cache=morpho_cache)))

    def _requires_morpho_mapping(self):
        return self._morphology_labels is not None and self.count_morphologies()

    def _morpho_backmap(self, locs):
        locs = locs.copy()
        cols = locs[:, 1:]
        ign_b = cols[:, 0] == -1
        ign_p = cols[:, 1] == -1
        semi = ign_b != ign_p
        if np.any(semi):
            raise ValueError(
                f"Invalid data at {np.nonzero(semi)[0]}. -1 needs to occur in "
                "either none or both columns to make point neuron connections."
            )
        to_map = ~ign_b
        if np.any(locs[to_map, 1:] < 0):
            raise ValueError(
                f"Invalid data at {np.nonzero(locs[to_map, 1:] < 0)[0]}, "
                "negative values are not valid morphology locations."
            )
        locs[to_map] = self.load_morphologies()._mapback(locs[to_map])
        return locs


class MorphologyRepository(Interface, engine_key="morphologies"):
    @abc.abstractmethod
    def all(self):
        """
        Fetch all of the stored morphologies.

        :returns: List of the stored morphologies.
        :rtype: List[~bsb.storage.interfaces.StoredMorphology]
        """
        pass

    @abc.abstractmethod
    def select(self, *selectors):
        """
        Select stored morphologies.

        :param selectors: Any number of morphology selectors.
        :type selectors: List[bsb.morphologies.selector.MorphologySelector]
        :returns: All stored morphologies that match at least one selector.
        :rtype: List[~bsb.storage.interfaces.StoredMorphology]
        """
        pass

    @abc.abstractmethod
    def save(self, name, morphology, overwrite=False):
        """
        Store a morphology

        :param name: Key to store the morphology under.
        :type name: str
        :param morphology: Morphology to store
        :type morphology: bsb.morphologies.Morphology
        :param overwrite: Overwrite any stored morphology that already exists under that
          name
        :type overwrite: bool
        :returns: The stored morphology
        :rtype: ~bsb.storage.interfaces.StoredMorphology
        """
        pass

    @abc.abstractmethod
    def has(self, name):
        """
        Check whether a morphology under the given name exists

        :param name: Key of the stored morphology.
        :type name: str
        :returns: Whether the key exists in the repo.
        :rtype: bool
        """
        pass

    def __contains__(self, item):
        return self.has(item)

    @abc.abstractmethod
    def preload(self, name):
        """
        Load a stored morphology as a morphology loader.

        :param name: Key of the stored morphology.
        :type name: str
        :returns: The stored morphology
        :rtype: ~bsb.storage.interfaces.StoredMorphology
        """
        pass

    @abc.abstractmethod
    def load(self, name):
        """
        Load a stored morphology as a constructed morphology object.

        :param name: Key of the stored morphology.
        :type name: str
        :returns: A morphology
        :rtype: ~bsb.morphologies.Morphology
        """
        pass

    @abc.abstractmethod
    def get_meta(self, name):
        """
        Get the metadata of a stored morphology.

        :param name: Key of the stored morphology.
        :type name: str
        :returns: Metadata dictionary
        :rtype: dict
        """
        pass

    @abc.abstractmethod
    def get_all_meta(self):
        """
        Get the metadata of all stored morphologies.
        :returns: Metadata dictionary
        :rtype: dict
        """
        pass

    @abc.abstractmethod
    def set_all_meta(self, all_meta):
        """
        Set the metadata of all stored morphologies.
        :param all_meta: Metadata dictionary.
        :type all_meta: dict
        """
        pass

    @abc.abstractmethod
    def update_all_meta(self, meta):
        """
        Update the metadata of stored morphologies with the provided key values

        :param meta: Metadata dictionary.
        :type meta: str
        """
        pass

    def list(self):
        """
        List all the names of the morphologies in the repository.
        """
        return [loader.name for loader in self.all()]


class ConnectivitySet(Interface):
    """
    Stores the connections between 2 types of cell as ``local`` and ``global`` locations.
    A location is a cell id, referring to the n-th cell in the chunk, a branch id, and a
    point id, to specify the location on the morphology. Local locations refer to cells on
    this chunk, while global locations can come from any chunk and is associated to a
    certain chunk id as well.

    Locations are either placement-context or chunk dependent: You may form connections
    between the n-th cells of a placement set (using
    :meth:`~.storage.interfaces.ConnectivitySet.connect`), or of the n-th cells of 2
    chunks (using :meth:`~.storage.interfaces.ConnectivitySet.chunk_connect`).

    A cell has both incoming and outgoing connections; when speaking of incoming
    connections, the local locations are the postsynaptic cells, and when speaking of
    outgoing connections they are the presynaptic cells. Vice versa for the global
    connections.
    """

    # The following attributes must be set on each ConnectivitySet by the engine:
    tag: str
    pre_type_name: str
    post_type_name: str
    pre_type: "CellType"
    post_type: "CellType"

    @abc.abstractmethod
    def __len__(self):
        pass

    @classmethod
    @abc.abstractmethod
    def create(cls, engine, tag):
        """
        Must create the placement set.
        """
        pass

    @obj_str_insert
    def __repr__(self):
        cstr = f"with {len(self)} connections" if len(self) else "without connections"
        return f"'{self.tag}' {cstr}"

    @staticmethod
    @abc.abstractmethod
    def exists(engine, tag):
        """
        Must check the existence of the connectivity set
        """
        pass

    def require(self, engine, tag):
        """
        Must make sure the connectivity set exists. The default
        implementation uses the class's ``exists`` and ``create`` methods.
        """
        if not self.exists(engine, tag):
            self.create(engine, tag)

    @abc.abstractmethod
    def clear(self, chunks=None):
        """
        Must clear (some chunks of) the placement set
        """
        pass

    @classmethod
    @abc.abstractmethod
    def get_tags(cls, engine):
        """
        Must return the tags of all existing connectivity sets.

        :param engine: Storage engine to inspect.
        """
        pass

    @abc.abstractmethod
    def connect(self, pre_set, post_set, src_locs, dest_locs):
        """
        Must connect the ``src_locs`` to the ``dest_locs``, interpreting the cell ids
        (first column of the locs) as the cell rank in the placement set.
        """
        pass

    @abc.abstractmethod
    def chunk_connect(self, src_chunk, dst_chunk, src_locs, dst_locs):
        """
        Must connect the ``src_locs`` to the ``dest_locs``, interpreting the cell ids
        (first column of the locs) as the cell rank in the chunk.
        """
        pass

    @abc.abstractmethod
    def get_local_chunks(self, direction):
        """
        Must list all the local chunks that contain data in the given ``direction``
        (``"inc"`` or ``"out"``).
        """
        pass

    @abc.abstractmethod
    def get_global_chunks(self, direction, local_):
        """
        Must list all the global chunks that contain data coming from a ``local`` chunk
        in the given ``direction``
        """
        pass

    @abc.abstractmethod
    def nested_iter_connections(self, direction=None, local_=None, global_=None):
        """
        Must iterate over the connectivity data, leaving room for the end-user to set up
        nested for loops:

        .. code-block:: python

          for dir, itr in self.nested_iter_connections():
              for lchunk, itr in itr:
                  for gchunk, data in itr:
                      print(f"Nested {dir} block between {lchunk} and {gchunk}")

        If a keyword argument is given, that axis is not iterated over, and the amount of
        nested loops is reduced.
        """
        pass

    @abc.abstractmethod
    def flat_iter_connections(self, direction=None, local_=None, global_=None):
        """
        Must iterate over the connectivity data, yielding the direction, local chunk,
        global chunk, and data:

        .. code-block:: python

          for dir, lchunk, gchunk, data in self.flat_iter_connections():
              print(f"Flat {dir} block between {lchunk} and {gchunk}")

        If a keyword argument is given, that axis is not iterated over, and the value is
        fixed in each iteration.
        """
        pass

    @abc.abstractmethod
    def load_block_connections(self, direction, local_, global_):
        """
        Must load the connections from ``direction`` perspective between ``local_`` and
        ``global_``.

        :returns: The local and global connections locations
        :rtype: Tuple[numpy.ndarray, numpy.ndarray]
        """
        pass

    @abc.abstractmethod
    def load_local_connections(self, direction, local_):
        """
        Must load all the connections from ``direction`` perspective in ``local_``.

        :returns: The local connection locations, a vector of the global connection chunks
          (1 chunk id per connection), and the global connections locations. To identify a
          cell in the global connections, use the corresponding chunk id from the second
          return value.
        :rtype: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
        """
        pass

    def load_connections(self):
        """
        Loads connections as a ``CSIterator``.

        :returns: A connectivity set iterator, that will load data
        """
        return ConnectivityIterator(self, "out")


class ConnectivityIterator:
    def __init__(
        self, cs: ConnectivitySet, direction, lchunks=None, gchunks=None, scoped=True
    ):
        self._cs = cs
        self._dir = direction
        self._lchunks = lchunks
        self._scoped = scoped
        self._gchunks = gchunks

    def __copy__(self):
        lchunks = self._lchunks.copy() if self._lchunks is not None else None
        gchunks = self._gchunks.copy() if self._gchunks is not None else None
        return ConnectivityIterator(self._cs, self._dir, lchunks, gchunks)

    def __len__(self):
        return len(self.all()[0])

    def __iter__(self):
        """
        Iterate over the connection locations chunk by chunk.

        :returns: The presynaptic location matrix and postsynaptic location matrix.
        :rtype: Tuple[numpy.ndarray, numpy.ndarray]
        """
        for _, pre_locs, _, post_locs in self.chunk_iter():
            yield from zip(pre_locs, post_locs)

    def chunk_iter(self):
        """
        Iterate over the connection data chunk by chunk.

        :returns: The presynaptic chunk, presynaptic locations, postsynaptic chunk,
          and postsynaptic locations.
        :rtype: Tuple[~bsb.storage._chunks.Chunk, numpy.ndarray, ~bsb.storage._chunks.Chunk, numpy.ndarray]
        """
        yield from (
            self._offset_block(*data)
            for data in self._cs.flat_iter_connections(
                self._dir, self._lchunks, self._gchunks
            )
        )

    @immutable()
    def as_globals(self):
        self._scoped = False

    @immutable()
    def as_scoped(self):
        self._scoped = True

    @immutable()
    def outgoing(self):
        self._dir = "out"
        self._lchunks, self._gchunks = self._gchunks, self._lchunks

    @immutable()
    def incoming(self):
        self._dir = "inc"
        self._lchunks, self._gchunks = self._gchunks, self._lchunks

    @immutable()
    def to(self, chunks):
        if isinstance(chunks, Chunk) and chunks.ndim == 1:
            chunks = [chunks]
        if self._dir == "inc":
            self._lchunks = chunks
        else:
            self._gchunks = chunks

    @immutable()
    def from_(self, chunks):
        if isinstance(chunks, Chunk) and chunks.ndim == 1:
            chunks = [chunks]
        if self._dir == "out":
            self._lchunks = chunks
        else:
            self._gchunks = chunks

    def all(self):
        pre_blocks = []
        post_blocks = []
        lens = []
        for _, pre_block, _, post_block in self.chunk_iter():
            pre_blocks.append(pre_block)
            post_blocks.append(post_block)
            lens.append(len(pre_block))
        pre_locs = np.empty((sum(lens), 3), dtype=int)
        post_locs = np.empty((sum(lens), 3), dtype=int)
        ptr = 0
        for len_, pre_block, post_block in zip(lens, pre_blocks, post_blocks):
            pre_locs[ptr : ptr + len_] = pre_block
            post_locs[ptr : ptr + len_] = post_block
            ptr += len_
        return pre_locs, post_locs

    def _offset_block(self, direction: str, lchunk, gchunk, data):
        loff = self._local_chunk_offsets()
        goff = self._global_chunk_offsets()
        llocs, glocs = data
        llocs[:, 0] += loff[lchunk]
        glocs[:, 0] += goff[gchunk]
        if direction == "out":
            return lchunk, llocs, gchunk, glocs
        else:
            return gchunk, glocs, lchunk, llocs

    @functools.cache
    def _local_chunk_offsets(self):
        source = self._cs.post_type if self._dir == "inc" else self._cs.pre_type
        return self._chunk_offsets(source, self._lchunks)

    @functools.cache
    def _global_chunk_offsets(self):
        source = self._cs.pre_type if self._dir == "inc" else self._cs.post_type
        return self._chunk_offsets(source, self._gchunks)

    def _chunk_offsets(self, source, chunks):
        stats = source.get_placement_set().get_chunk_stats()
        if self._scoped and chunks is not None:
            stats = {chunk: item for chunk, item in stats.items() if int(chunk) in chunks}
        offsets = {}
        ctr = 0
        for chunk, len_ in sorted(
            stats.items(), key=lambda k: Chunk.from_id(int(k[0]), None).id
        ):
            offsets[Chunk.from_id(int(chunk), None)] = ctr
            ctr += len_
        return offsets


class StoredMorphology:
    def __init__(self, name, loader, meta):
        self.name = name
        self._loader = loader
        self._meta = meta

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def get_meta(self):
        return self._meta.copy()

    def load(self):
        return self._loader()

    def cached_load(self, labels=None):
        if labels is not None:
            labels = tuple(labels)
        return self._cached_load(labels)

    @functools.cache
    def _cached_load(self, labels):
        return self.load().set_label_filter(labels).as_filtered()


class GeneratedMorphology(StoredMorphology):
    def __init__(self, name, generated, meta):
        super().__init__(name, lambda: generated, meta)


__all__ = [
    "ConnectivityIterator",
    "ConnectivitySet",
    "Engine",
    "FileStore",
    "GeneratedMorphology",
    "Interface",
    "MorphologyRepository",
    "NetworkDescription",
    "NoopLock",
    "PlacementSet",
    "ReadOnlyManager",
    "StorageNode",
    "StoredFile",
    "StoredMorphology",
]



"""
Module for binary space partitioning, to facilitate optimal runtime complexity for n-point
problems.
"""

import abc

from rtree import index as rtree


class BoxTreeInterface(abc.ABC):
    """
    Tree for fast lookup of queries of axis aligned rhomboids.
    """

    @abc.abstractmethod
    def query(self, boxes, unique=False):
        """
        Should return a generator that yields lists of intersecting IDs per query box if
        ``unique=False``. If ``unique=True``, yield a flat list of unique intersecting box
        IDs for all queried boxes.
        """
        pass

    @abc.abstractmethod
    def __len__(self):
        pass


class _BoxRTree(BoxTreeInterface):
    """
    Tree for fast lookup of queries of axis aligned rhomboids using the Rtree package.
    """

    def __init__(self, boxes):
        self._rtree = rtree.Index(properties=rtree.Property(dimension=3))
        for id, box in enumerate(boxes):
            self._rtree.insert(id, box)

    def __len__(self):
        return len(self._rtree)

    def query(self, boxes, unique=False):
        """
        Given an iterable of ``(min_x, min_y, min_z, max_x, max_y, max_z)`` box tuples,
        find all the boxes that intersect with them.

        :param boxes: Boxes to look for intersections with.
        :type boxes: Iterable[Tuple[float, float, float, float, float, float]]
        :param unique: If ``True``, return a flat generator of unique results. If ``False``
            (default), per box in ``boxes``, return a list of intersecting boxes.
        :returns: See ``unique``.
        :rtype: Union[Iterator[List[Tuple[float, float, float, float, float, float]]],
            Iterator[Tuple[float, float, float, float, float, float]]]
        """
        all_ = (list(self._rtree.intersection(box, objects=False)) for box in boxes)
        if unique:
            seen = set()
            # Double for loop over results, skipping those that have been seen before.
            yield from (
                seen.add(elem) or elem for arr in all_ for elem in arr if elem not in seen
            )
        else:
            yield from all_


# Cheapo provider pattern.
class BoxTree(_BoxRTree):
    """
    Tree for fast lookup of repeat queries of axis aligned rhomboids.
    """

    pass


__all__ = ["BoxTreeInterface", "BoxTree"]



"""
    This module contains shorthand ``reference`` definitions. References are used in the
    configuration module to point to other locations in the Configuration object.

    Minimally a reference is a function that takes the configuration root and the current
    node as arguments, and returns another node in the configuration object::

      def some_reference(root, here):
          return root.other.place

    More advanced usage of references will include custom reference errors.
"""


class Reference:  # pragma: nocover
    def __call__(self, root, here):
        return here

    def up(self, here, to=None):
        if to is None:
            return here._config_parent
        while not isinstance(here, to):
            try:
                here = here._config_parent
            except AttributeError:
                return None
        return here


class CellTypeReference(Reference):
    def __call__(self, root, here):
        return root.cell_types

    def is_ref(self, value):
        from ..cell_types import CellType

        return isinstance(value, CellType)


class ConnectionTypeReference(Reference):
    def __call__(self, root, here):
        return root.connectivity

    def is_ref(self, value):
        from ..connectivity import ConnectionStrategy

        return isinstance(value, ConnectionStrategy)


class PartitionReference(Reference):
    def __call__(self, root, here):
        return root.partitions

    def is_ref(self, value):
        from ..topology import Partition

        return isinstance(value, Partition)


class RegionReference(Reference):
    def __call__(self, root, here):
        return root.regions

    def is_ref(self, value):
        from ..topology import Region

        return isinstance(value, Region)


class RegionalReference(Reference):
    def __call__(self, root, here):
        merged = root.regions.copy()
        merged.update(root.partitions)
        return merged

    def is_ref(self, value):
        from ..topology import Partition, Region

        return isinstance(value, Region) or isinstance(value, Partition)


class PlacementReference(Reference):
    def __call__(self, root, here):
        return root.placement

    def is_ref(self, value):
        from ..placement import PlacementStrategy

        return isinstance(value, PlacementStrategy)


class ConnectivityReference(Reference):
    def __call__(self, root, here):
        return root.connectivity

    def is_ref(self, value):
        from ..connectivity import ConnectionStrategy

        return isinstance(value, ConnectionStrategy)


class SimCellModelReference(Reference):
    def __call__(self, root, here):
        from ..simulation.simulation import Simulation

        sim = self.up(here, Simulation)
        return sim.cell_models

    def is_ref(self, value):
        from ..simulation.cell import CellModel

        return isinstance(value, CellModel)


cell_type_ref = CellTypeReference()
conn_type_ref = ConnectionTypeReference()
partition_ref = PartitionReference()
placement_ref = PlacementReference()
connectivity_ref = ConnectivityReference()
regional_ref = RegionalReference()
region_ref = RegionReference()
sim_cell_model_ref = SimCellModelReference()

__all__ = [
    "Reference",
    "cell_type_ref",
    "conn_type_ref",
    "partition_ref",
    "placement_ref",
    "connectivity_ref",
    "regional_ref",
    "region_ref",
    "sim_cell_model_ref",
]
__api__ = ["Reference"]



# isort: off
# Load module before others to prevent partially initialized modules
from .strategy import PlacementStrategy

# isort: on
from .arrays import ParallelArrayPlacement
from .import_ import CsvImportPlacement, ImportPlacement
from .random import RandomPlacement
from .strategy import Entities, FixedPositions



"""
This module contains the classes required to construct options.
"""

import argparse
import functools
import os
import pathlib

import toml

from .exceptions import OptionError
from .reporting import warn


class OptionDescriptor:
    """
    Base option property descriptor. Can be inherited from to create a cascading property
    such as the default CLI, env & script descriptors.
    """

    def __init_subclass__(cls, slug=None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.slug = slug

    def __init__(self, *tags):
        self.tags = tags

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, f"_bsbopt_{self.slug}_value", instance.get_default())

    def __set__(self, instance, value):
        set_value = getattr(instance, "setter", lambda x: x)(value)
        setattr(instance, f"_bsbopt_{self.slug}_value", set_value)

    def __delete__(self, instance):
        try:
            delattr(instance, f"_bsbopt_{self.slug}_value")
        except AttributeError:
            pass

    def is_set(self, instance):
        return hasattr(instance, f"_bsbopt_{self.slug}_value")


class CLIOptionDescriptor(OptionDescriptor, slug="cli"):
    """
    Descriptor that retrieves its value from the given CLI command arguments.
    """

    pass


class EnvOptionDescriptor(OptionDescriptor, slug="env"):
    """
    Descriptor that retrieves its value from the environment variables.
    """

    def __init__(self, *args, flag=False):
        super().__init__(*args)
        self.flag = flag

    def __get__(self, instance, owner):
        if instance is None:
            return self
        getter = getattr(instance, "getter", lambda x: x)
        # Iterate the env for all tags, if none are set this returns `None`
        for tag in self.tags:
            if tag in os.environ:
                return getter(self._parse(os.environ[tag]))

    def __set__(self, instance, value):
        value = getattr(instance, "setter", lambda x: x)(value)
        parsed = self._rev_parse(value)
        for tag in self.tags:
            os.environ[tag] = parsed

    def __delete__(self, instance):
        for tag in self.tags:
            try:
                del os.environ[tag]
            except KeyError:
                pass

    def is_set(self, instance):
        return any(tag in os.environ for tag in self.tags)

    def _parse(self, value):
        if self.flag:
            if value is True or str(value).strip().upper() in ("ON", "TRUE", "1", "YES"):
                return True
            else:
                return False
        else:
            return value

    def _rev_parse(self, value):
        if self.flag:
            return "ON" if value else "OFF"
        else:
            return str(value)


class ScriptOptionDescriptor(OptionDescriptor, slug="script"):
    """
    Descriptor that retrieves and sets its value from/to the :mod:`bsb.options` module.
    """

    # This class uses `self.tags[0]`, because all tags are aliases of eachother in the
    # options module.

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if not self.tags:
            # This option has no options module binding
            return None
        from .options import get_module_option

        return get_module_option(self.tags[0])

    def __set__(self, instance, value):
        if not self.tags:
            # This option has no options module binding
            raise OptionError(f"{instance.name} can't be set through `bsb.options`.")
        from .options import set_module_option

        set_value = getattr(instance, "setter", lambda x: x)(value)
        return set_module_option(self.tags[0], set_value)

    def __delete__(self, instance):
        from .options import reset_module_option

        for tag in self.tags:
            reset_module_option(tag)

    def is_set(self, instance):
        from .options import is_module_option_set

        return any(is_module_option_set(tag) for tag in self.tags)


class ProjectOptionDescriptor(OptionDescriptor, slug="project"):
    """
    Descriptor that retrieves and stores values in the `pyproject.toml` file. Traverses
    up the filesystem tree until one is found.
    """

    def __init__(self, *tags):
        if len(tags) > 1:  # pragma: nocover
            raise OptionError(f"Project option can have only 1 tag, got {tags}.")
        super().__init__(*(tags[0].split(".") if tags else ()))

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.tags:
            _, proj = _pyproject_bsb()
            for tag in self.tags[:-1]:
                proj = proj.get(tag, None)
                if proj is None:
                    return None
            return proj.get(self.tags[-1], None)

    def __set__(self, instance, value):
        if self.tags:
            path, proj = _pyproject_bsb()
            deeper = proj
            for tag in self.tags[:-1]:
                deeper = deeper.setdefault(tag, {})
            deeper[self.tags[-1]] = value
            _save_pyproject_bsb(proj)

    def __delete__(self, instance):
        if self.tags:
            path, proj = _pyproject_bsb()
            for tag in self.tags[:-1]:
                proj = proj.get(tag, None)
                if proj is None:
                    return None
            try:
                del proj[self.tags[-1]]
            except KeyError:
                pass
            else:
                _save_pyproject_bsb(proj)

    def is_set(self, instance):
        if self.tags:
            _, proj = _pyproject_bsb()
            for tag in self.tags[:-1]:
                proj = proj.get(tag, None)
                if proj is None:
                    return False
            return self.tags[-1] in proj
        else:
            return False


class BsbOption:
    """
    Base option class. Can be subclassed to create new options.
    """

    def __init__(self, positional=False):
        self.positional = positional

    def __init_subclass__(
        cls,
        name=None,
        env=(),
        project=(),
        cli=(),
        script=(),
        description=None,
        flag=False,
        inverted=False,
        list=False,
        readonly=False,
        action=False,
    ):
        """
        Subclass hook that defines the characteristics of the subclassed option class.

        :param name: Unique name for identification
        :type name: str
        :param cli: Positional arguments for the :class:`.CLIOptionDescriptor` constructor.
        :type cli: iterable
        :param cli: Positional arguments for the :class:`.CLIOptionDescriptor` constructor.
        :type cli: iterable
        :param env: Positional arguments for the :class:`.EnvOptionDescriptor` constructor.
        :type env: iterable
        :param script: Positional arguments for the :class:`.ScriptOptionDescriptor` constructor.
        :type script: iterable
        :param description: Description of the option's purpose for the user.
        :type description: str
        :param flag: Indicates that the option is a flag and should toggle on a default off boolean when given.
        :type flag: boolean
        :param inverted: Used only for flags. Indicates that the flag is default on and is toggled off when given.
        :param list: Indicates that the option takes multiple values.
        :type list: boolean
        :param readonly: Indicates that an option can be accessed but not be altered from the ``bsb.options`` module.
        :type readonly: boolean
        :param action: Indicates that the option should execute its ``action`` method.
        :type action: boolean
        """
        if name is None:
            raise OptionError("Options must be given a name in the class argument list.")
        cls.name = name
        cls.env = EnvOptionDescriptor(*env, flag=flag)
        cls.project = ProjectOptionDescriptor(*project)
        cls.cli = CLIOptionDescriptor(*cli)
        cls.script = ScriptOptionDescriptor(*script)
        cls.description = description
        cls.is_flag = flag
        cls.inverted_flag = inverted
        cls.use_extend = list
        cls.readonly = readonly
        cls.use_action = action
        cls.positional = False

    def get(self, prio=None):
        """
        Get the option's value. Cascades the script, cli, env & default descriptors together.

        :returns: option value
        """
        try:
            if prio is not None:
                return getattr(self, prio)

            cls = self.__class__
            if cls.script.is_set(self):
                return self.script
            if cls.cli.is_set(self):
                return self.cli
            if cls.project.is_set(self):
                return self.project
            if cls.env.is_set(self):
                return self.env
            return self.get_default()
        except Exception as e:
            warn(f"Error retrieving option '{self.name}'.", log_exc=e)
            return self.get_default()

    def is_set(self, slug):
        if descriptor := getattr(type(self), slug, None):
            return descriptor.is_set(self)
        else:
            return False

    def get_default(self):
        """
        Override to specify the default value of the option.
        """
        return None

    def get_cli_tags(self):
        """
        Return the ``argparse`` positional arguments from the tags.

        :returns: ``-x`` or ``--xxx`` for each CLI tag.
        :rtype: list
        """
        if self.positional:
            longest = ""
            for t in type(self).cli.tags:
                if len(t) >= len(longest):
                    longest = t
            return [longest]
        else:
            return [("--" if len(t) != 1 else "-") + t for t in type(self).cli.tags]

    def add_to_parser(self, parser, level):
        """
        Register this option into an ``argparse`` parser.
        """
        if not self.get_cli_tags():
            return

        kwargs = {}
        kwargs["help"] = self.description
        kwargs["dest"] = level * "_" + self.name
        kwargs["action"] = "store"
        if self.positional:
            kwargs["nargs"] = "?"
            kwargs["metavar"] = self.get_cli_tags()
            args = []
        else:
            args = self.get_cli_tags()
        if self.is_flag:
            kwargs["action"] += "_false" if self.inverted_flag else "_true"
            kwargs["default"] = argparse.SUPPRESS
        if self.use_extend:
            kwargs["action"] = "extend"
            kwargs["nargs"] = "+"
        if self.use_action:
            kwargs["dest"] = "internal_action_list"
            kwargs["action"] = "append_const"
            kwargs["const"] = self.action

        parser.add_argument(*args, **kwargs)

    @classmethod
    def register(cls):
        """
        Register this option class into the :mod:`bsb.options` module.
        """
        from . import options

        opt = cls()
        options.register_option(cls.name, opt)
        return opt

    def unregister(self):
        """
        Remove this option class from the :mod:`bsb.options` module, not part of the
        public API as removing options is undefined behavior but useful for testing.
        """
        from . import options

        options.unregister_option(self)


@functools.cache
def _pyproject_path():
    path = pathlib.Path.cwd()
    while str(path)[len(path.drive) :] != path.root:
        proj = path / "pyproject.toml"
        if proj.exists():
            return proj
        path = path.parent


def _pyproject_content():
    path = _pyproject_path()
    if path:
        with open(path, "r") as f:
            return path.resolve(), toml.load(f)
    else:
        return None, {}  # pragma: nocover


def _pyproject_bsb():
    path, content = _pyproject_content()
    return path, content.get("tools", {}).get("bsb", {})


def _save_pyproject_bsb(project):
    path, content = _pyproject_content()
    if path is None:
        raise OptionError(
            "No 'pyproject.toml' in current dir or parents,"
            + " can't set project settings."
        )
    content.setdefault("tools", {})["bsb"] = project
    with open(path, "w") as f:
        toml.dump(content, f)


__all__ = [
    "BsbOption",
    "CLIOptionDescriptor",
    "EnvOptionDescriptor",
    "OptionDescriptor",
    "ProjectOptionDescriptor",
    "ScriptOptionDescriptor",
]
